

"""Defines widgets which may be used as page footers"""

from string import Template
from .. import skiboot
from .. import tag
from . import Widget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class SimpleFooter(Widget):
    """Defines a div, with a paragraph of text"""

    error_location = (0,0,0)

    arg_descriptions = {
                        'footer_text':FieldArg("text",'', jsonset=True),
                        'para_class':FieldArg("cssclass",""),
                        'error_class':FieldArg("cssclass", ""),
                        'paradiv_class':FieldArg("cssclass",""),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
         footer_text: The text to appear in the footer
         para_class: css class of the footer text paragraph
         error_class: css class of the error paragraph
        """

        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # hidden error
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="div")
        self[1][0] = tag.Part(tag_name="p")

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the paragraph"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('paradiv_class'):
            self[1].update_attribs({"class":self.get_field_value('paradiv_class')})
        if self.get_field_value('para_class'):
            self[1][0].update_attribs({"class":self.get_field_value('para_class')})
        if self.get_field_value("footer_text"):
            self[1][0][0] = self.get_field_value("footer_text")
        # set an id in the footer_text paragraph
        self[1][0].insert_id()

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """stores ident in footer text paragraph"""
        return self._make_fieldvalues(textident = self[1][0].get_id()) # ident of the footer_text paragraph

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p>       <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- with class paradiv_class -->
    <p>  <!-- with class para_class -->
    <!-- footer_text appears in this paragraph -->
    </p>
  </div>
</div>
"""






import re, json, os
# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w_]')

from skipole.skilift import fromjson, off_piste, set_label, versions, lib_list, labels
from skipole import FailPage, ValidateError, ServerError, GoTo

from .. import utils, css_styles


def retrieve_operations_data(skicall):
    "Retrieves field data for operations page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data, but keep status
    # so the status message is displayed after any operations
    utils.clear_call_data(call_data, keep='status')

    editedprojname = call_data['editedprojname']

    # project background color
    backcol = fromjson.get_defaults(editedprojname, key="backcol")
    if backcol:
        page_data['htmlbackcol','input_text'] = backcol

    # project body class
    body_class = fromjson.get_defaults(editedprojname, key="body_class")
    if body_class:
        page_data['bodyclass','input_text'] = body_class

    # get default css links for widget css_defaults
    css_list = fromjson.get_defaults(editedprojname, key='css_links')

    if css_list:
        contents = []
        for label in css_list:
            row = [label,
                   label, label, label, '',
                   True,
                   True,
                   True,
                   False]
            contents.append(row)
        # remove up arrow in row 0
        contents[0][1] = ''
        contents[0][5] = False
        # remove down arrow in row -1
        contents[-1][2] = ''
        contents[-1][6] = False
        page_data[('css_links', 'contents')] = contents


  #      contents: col 0 is the text to place in the first column,
  #                  col 1, 2, 3, 4 are the get field contents of links 1,2,3 and 4
  #                  col 5 - True if the first button and link is to be shown, False if not
  #                  col 6 - True if the second button and link is to be shown, False if not
  #                  col 7 - True if the third button and link is to be shown, False if not
  #                  col 8 - True if the fourth button and link is to be shown, False if not


    # get default js links for widget js_defaults
    js_list = fromjson.get_defaults(editedprojname, key='js_links')

    if js_list:
        contents = []
        for label in js_list:
            row = [label,
                   label, label, label, '',
                   True,
                   True,
                   True,
                   False]
            contents.append(row)
        # remove up arrow in row 0
        contents[0][1] = ''
        contents[0][5] = False
        # remove down arrow in row -1
        contents[-1][2] = ''
        contents[-1][6] = False
        page_data[('js_links', 'contents')] = contents


def handle_upload(skicall):
    "handle uploaded defaults.json file"

    call_data = skicall.call_data
    page_data = skicall.page_data
    editedprojname = call_data['editedprojname']

    # get uploaded file contents
    if "upload" not in call_data:
        raise FailPage("upload missing from call data")
    file_contents = call_data["upload"]
    if not file_contents:
        raise FailPage("No data found in file")
    defaults_dict = json.loads(file_contents.decode())
    fromjson.save_defaults(editedprojname, defaults_dict)
    call_data['status'] = 'Defaults file installed'


def set_download(skicall):
    "Set the path to the file to be downloaded"
    skicall.page_data['filepath'] = os.path.join(skicall.call_data['editedprojname'], "data", "defaults.json")


def set_widgets_css(skicall):
    "sets default css classes into widgets"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    off_piste.set_widget_css_to_default(editedprojname)
    call_data['status'] = 'Widget CSS classes set'


def submit_project_color(skicall):
    "set project default background color"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        backcol = call_data['htmlbackcol','input_text']
    except Exception:
        raise ValidateError(message='Invalid call')
    try:
        editedprojname = call_data['editedprojname']
        # set background color in existing pages
        off_piste.set_backcol_in_pages(editedprojname, backcol)
        # set the project colour for new pages
        fromjson.set_defaults(editedprojname, 'backcol', backcol)
    except (ValidateError, ServerError) as e:
        raise FailPage(message = e.message, widget = "htmlbackcol")
    call_data['status'] = 'Project background colour set'


def set_bodyclass_in_pages(skicall):
    "sets css class into page body tags"

    call_data = skicall.call_data
    page_data = skicall.page_data

    try:
        bodyclass = call_data['bodyclass','input_text']
    except Exception:
        raise ValidateError(message='Invalid call')
    # set body class in all project template pages
    editedprojname = call_data['editedprojname']
    off_piste.set_bodyclass_in_pages(editedprojname, bodyclass)
    # set the body class for new pages
    fromjson.set_defaults(editedprojname, 'body_class', bodyclass)
    if bodyclass:
        call_data['status'] = 'CSS class set in body tags'
    else:
        call_data['status'] = 'CSS class removed from body tags'


def add_default_css(skicall):
    "Add a label to the default css list"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    # check label is valid
    if not "new_css_label" in call_data:
        raise FailPage(message = "label missing")

    label = call_data["new_css_label"]
    if not label:
        raise FailPage(message = "label missing")

    if _AN.search(label):
        raise FailPage(message = "The label can only contain A-Z, a-z, 0-9 and the underscore character.")

    if '_' in label:
        labelparts = label.split('_')
        if labelparts[0].isalnum() and labelparts[1].isdigit():
            raise FailPage(message = "Invalid label (Danger of confusion with a page ident).")

    if label.isdigit():
        raise FailPage(message = "Invalid label (Danger of confusion with a page ident).")

    css_list = fromjson.get_defaults(editedprojname, key='css_links')
    if label in css_list:
        raise FailPage(message = "This label is already in the list")
    css_list.append(label)
    try:
        fromjson.set_defaults(editedprojname, key='css_links', value=css_list)
    except e:
        raise FailPage(message = "Unable to save defaults.json")


def css_remove(skicall):
    "Removes css default label"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    if 'css_label' in call_data:
        label = call_data['css_label']
        d_list = fromjson.get_defaults(editedprojname, 'css_links')
        if not label in d_list:
            return
        index = d_list.index(label)
        # remove
        del d_list[index]
        # save d_list
        try:
            fromjson.set_defaults(editedprojname, key='css_links', value=d_list)
        except e:
            raise FailPage(message = "Unable to save defaults.json")


def css_up(skicall):
    "Moves css default label up"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    if 'css_label' in call_data:
        label = call_data['css_label']
        d_list = fromjson.get_defaults(editedprojname, 'css_links')
        if not label in d_list:
            return
        index = d_list.index(label)
        # move up
        if not index:
            return
        d_list.insert(index-1, d_list.pop(index))
        # save d_list
        try:
            fromjson.set_defaults(editedprojname, key='css_links', value=d_list)
        except e:
            raise FailPage(message = "Unable to save defaults.json")


def css_down(skicall):
    "Moves css default label down"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    if 'css_label' in call_data:
        label = call_data['css_label']
        d_list = fromjson.get_defaults(editedprojname, 'css_links')
        if not label in d_list:
            return
        index = d_list.index(label)
        # move down
        if index == len(d_list)-1:
            return
        d_list.insert(index+1, d_list.pop(index))
        # save d_list
        try:
            fromjson.set_defaults(editedprojname, key='css_links', value=d_list)
        except e:
            raise FailPage(message = "Unable to save defaults.json")


def add_default_js(skicall):
    "Add a label to the default javascript list"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    # check label is valid
    if not "new_js_label" in call_data:
        raise FailPage(message = "label missing")

    label = call_data["new_js_label"]
    if not label:
        raise FailPage(message = "label missing")

    if _AN.search(label):
        raise FailPage(message = "The label can only contain A-Z, a-z, 0-9 and the underscore character.")

    if '_' in label:
        labelparts = label.split('_')
        if labelparts[0].isalnum() and labelparts[1].isdigit():
            raise FailPage(message = "Invalid label (Danger of confusion with a page ident).")

    if label.isdigit():
        raise FailPage(message = "Invalid label (Danger of confusion with a page ident).")

    js_list = fromjson.get_defaults(editedprojname, key='js_links')
    if label in js_list:
        raise FailPage(message = "This label is already in the list")
    js_list. append(label)
    try:
        fromjson.set_defaults(editedprojname, key='js_links', value=js_list)
    except e:
        raise FailPage(message = "Unable to save defaults.json")


def js_remove(skicall):
    "Removes js default label"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    if 'js_label' in call_data:
        label = call_data['js_label']
        d_list = fromjson.get_defaults(editedprojname, 'js_links')
        if not label in d_list:
            return
        index = d_list.index(label)
        # remove
        del d_list[index]
        # save d_list
        try:
            fromjson.set_defaults(editedprojname, key='js_links', value=d_list)
        except e:
            raise FailPage(message = "Unable to save defaults.json")


def js_up(skicall):
    "Moves js default label up"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    if 'js_label' in call_data:
        label = call_data['js_label']
        d_list = fromjson.get_defaults(editedprojname, 'js_links')
        if not label in d_list:
            return
        index = d_list.index(label)
        # move up
        if not index:
            return
        d_list.insert(index-1, d_list.pop(index))
        # save d_list
        try:
            fromjson.set_defaults(editedprojname, key='js_links', value=d_list)
        except e:
            raise FailPage(message = "Unable to save defaults.json")


def js_down(skicall):
    "Moves js default label down"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    if 'js_label' in call_data:
        label = call_data['js_label']
        d_list = fromjson.get_defaults(editedprojname, 'js_links')
        if not label in d_list:
            return
        index = d_list.index(label)
        # move down
        if index == len(d_list)-1:
            return
        d_list.insert(index+1, d_list.pop(index))
        # save d_list
        try:
            fromjson.set_defaults(editedprojname, key='js_links', value=d_list)
        except e:
            raise FailPage(message = "Unable to save defaults.json")


def set_javascript_to_skis(skicall):
    "Sets javascript labels to skis"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']
    liblabels = lib_list()
    projlabels = labels(editedprojname)
    try:
        for label in liblabels:
            # label is a widget or validator javascript label, of the format "ski_basic" etc
            # which needs to be given a value of the format "skis,ski_basic"
            set_label(editedprojname, label, "skis," + label)
        if "w3_css" in projlabels:
            if projlabels["w3_css"] == "https://www.w3schools.com/w3css/4/w3.css":
                set_label(editedprojname, "w3_css", "skis,w3_css")
    except ServerError as e:
        raise FailPage(message = e.message)
    call_data['status'] = 'Special file labels now point to the skis sub project'


def set_javascript_to_cdn(skicall):
    "Sets javascript labels to cdn"

    call_data = skicall.call_data
    page_data = skicall.page_data

    editedprojname = call_data['editedprojname']

    skipole_version, project_version = versions(editedprojname)
    # get this to point to https://cdn.jsdelivr.net/gh/bernie-skipole/skipole@5/skipole/templates/skis/static/js/
    # so primary version number is used
    version_numbers = skipole_version.split(".")
    starthttp =  "https://cdn.jsdelivr.net/gh/bernie-skipole/skipole@" + version_numbers[0] + "/skipole/templates/skis/static/js/"
    liblabels = lib_list()
    projlabels = labels(editedprojname)
    try:
        for label in liblabels:
            if label == "jquery_core":
                if projlabels[label] == "skis,jquery_core":
                    set_label(editedprojname, label, "https://cdn.jsdelivr.net/gh/jquery/jquery@3/dist/jquery.min.js")
                continue
            if label == "skipole_js":
                set_label(editedprojname, "skipole_js", starthttp +"skipole.min.js")
                continue
            # so label is a widget or validator javascript label, of the format "ski_basic" etc
            # which needs to be given a value of the format "basic.min.js"
            set_label(editedprojname, label, starthttp +label[4:]+".min.js")
        if "w3_css" in projlabels:
            if projlabels["w3_css"] == "skis,w3_css":
                set_label(editedprojname, "w3_css", "https://www.w3schools.com/w3css/4/w3.css")
    except ServerError as e:
        raise FailPage(message = e.message)
    call_data['status'] = "Special file labels now point to CDN's"







import html

from skipole import FailPage, ServerError
from skipole import skilift


def clear_call_data(call_data, keep=None):
    "Clears call data apart from set of required values, and anything in the keep list"
    required = ['editedprojname',
                'editedprojurl',
                'editedprojversion',
                'editedprojbrief',
                'adminproj',
                'extend_nav_buttons',
                'caller_ident']
    if keep:
        if isinstance(keep, str) and (keep not in required):
            required.append(keep)
        else:
            # presume keep is a list of items
            for item in keep:
                if item not in required:
                    required.append(item)
    temp_storage = {key:value for key,value in call_data.items() if key in required}
    call_data.clear()
    call_data.update(temp_storage)







"""
This module defines the classes:

Part which holds an html tag, with attributes and contents.

ClosedPart which is also a tag, with name and attributes, but no contents, such as <tag_name att="attribute" />.

Widget which inherits from Part and can be subclassed to create widgets
"""

import html, copy, uuid

from collections import OrderedDict

from urllib.parse import quote, quote_plus

from . import skiboot
from .excepts import ValidateError, ServerError


def expand_text(text, escape=True, linebreaks=True, replace_strings=[]):
    "Sets any replace string, expands any newlines in text to linebreaks, and applies html.escape to text"
    if not text:
        return ""
    if not isinstance(text, str):
        text = str(text)
    if replace_strings:
        try:
            newtext = text % tuple(replace_strings)
        except Exception:
            # unable to complete the replace, so carry on
            pass
        else:
            # no errors, set text to newtext
            text = newtext
    if escape:
        text = html.escape(text)
    if not linebreaks:
        return text
    # linebreaks must be true
    return "<br />".join(text.split("\n"))


def make_dictionary(input_dict, proj_ident):
    """Used by other functions, takes input_dict and returns
       an ordered dict, with TextBlocks, idents converted for json compatability"""
    output_dict = OrderedDict()
    if not input_dict: return output_dict
    for key, val in input_dict.items():
        if val is None:
            output_dict[key] = None
        elif val == '':
            output_dict[key] = ""
        elif isinstance(val, list):
            output_dict[key] = make_list(val, proj_ident)
        elif isinstance(val, dict):
            output_dict[key] = make_dictionary(val, proj_ident)
        elif val is True:
            output_dict[key] = True
        elif val is False:
            output_dict[key] = False
        elif isinstance(val, TextBlock):
            output_dict[key] = val.textref
        elif isinstance(val, skiboot.Ident):
            if val.proj == proj_ident:
                # ident is this project, put the number only
                output_dict[key] = val.num
            else:
                # ident is another project, put the full ident
                output_dict[key] = [val.proj, val.num]
        else:
            output_dict[key] = str(val)
    return output_dict


def make_list(input_list, proj_ident):
    """Used by other functions, takes input_list and returns a list with items converted"""
    if not input_list: return []
    output_list = []
    for item in input_list:
        if item is None:
            output_list.append(None)
        elif item == '':
            output_list.append('')
        elif isinstance(item, list):
            output_list.append(make_list(item, proj_ident))
        elif isinstance(item, dict):
            output_list.append(make_dictionary(item, proj_ident))
        elif item is True:
            output_list.append(True)
        elif item is False:
            output_list.append(False)
        elif isinstance(item, skiboot.Ident):
            if item.proj == proj_ident:
                output_list.append(item.num)
            else:
                # ident is another project, put the full ident
                output_list.append([item.proj, item.num])
        else:
            output_list.append(str(item))
    return output_list


class ParentPart(object):
    "The parent of Part and Closedpart objects"

    def __init__(self, tag_name, attribs, show, brief):
        # self.name is only set if a part is also a widget
        self.name=None

        self.tag_name = tag_name
        self._attribs = {}
        self.hide_if_empty = False

        # set when set_idents is called on the containing page
        self.ident_string = ''


        # self.brief is a string describing what this part does
        self.brief = brief

        if attribs:
            self.set_attribs(attribs)

        # set show to False if this part is not to be shown
        self.show = show
        self._error = ""

        # embedded is a tuple of section name, parent widget name, parent widget container integer
        # If not within a section, section name is an empty string.
        # If not in a widget, the widget name is empty string and container integer is set to None
        # if not in a container, the container integer is None
        self.embedded = ('','', None)

        # If this is in a section, this is set
        self.placename = ''

        # parts contains a list of sub parts, always empty for a closedpart
        self.parts = []

    def set_idents(self, ident_string, widgets, section_places, embedded=('','',None)):
        """Sets self.ident_string in this and sub parts, called when a page is saved,
             embedded is (section_name, parent_widget_name, parent_widget_container)"""
        self.ident_string = ident_string
        parent_container = None
        # is this part in a widget, and container not set
        if embedded[1] and (embedded[2] is None):
            # get the widget
            parent_widget = widgets[embedded[1]]
            # check the parent_widget containers
            for cont in range(parent_widget.len_containers()):
                parent_container_string = parent_widget.get_container_string_loc(cont)
                if parent_container_string:
                    container_ident_string = parent_widget.ident_string + '-' + parent_container_string
                else:
                    container_ident_string = parent_widget.ident_string
                if self.ident_string.startswith(container_ident_string):
                    parent_container = cont
                    break
        if parent_container is None:
            self.embedded = embedded
        else:
            self.embedded = (embedded[0], embedded[1], parent_container)
        if self.name:
            # parts beneath this named widget will be within this widget
            embedded_parts = (embedded[0], self.name, None)
            # add the widget to the widgets dictionary
            widgets[self.name] = self
        else:
            embedded_parts = self.embedded
        for index, part in enumerate(self.parts):
            if hasattr(part, "set_idents"):
                part_ident_string = ident_string + "-" + str(index)
                part.set_idents(part_ident_string, widgets, section_places, embedded_parts)


    def set_unique_names(self, name_list):
        "Ensures this item, and sub items have a unique name"
        if self.name:
            number = 2
            new_name = self.name
            while new_name in name_list:
                new_name = self.name + str(number)
                number += 1
            self.name = new_name 
            name_list.append(self.name)
        for index, part in enumerate(self.parts):
            if hasattr(part, "set_unique_names"):
                part.set_unique_names(name_list)


    @property
    def proj_ident(self):
        if not self.ident_string:
            return ''
        return self.ident_string.split('_')[0]

    def insert_id(self, id_string=''):
        """Adds the id_string to the part attributes. If no id_string given
           but the part has a name, sets the id to the part name, otherwise
           sets the id to self.ident_string"""
        if id_string:
            self.update_attribs({"id": id_string})
            return
        if self.name:
            # This only applies to named widgets
            if self.placename:
                full_id = self.placename + '-' + self.name
                self.update_attribs({"id": full_id})
            else:
                self.update_attribs({"id": self.name})
            return
        if self.ident_string:
            self.update_attribs({"id": self.ident_string})

    def get_id(self):
        return self.get_attrib_value('id')

    def get_class(self):
        return self.get_attrib_value('class')

    def set_class(self, value):
        if (not value) and ('class' in self._attribs):
            del self._attribs['class']
        else:
            self._attribs['class'] = value

    def set_hide(self, value=True):
        "If True, Sets display:none !important into style, if False removes it"
        if value:
            attribs = self._attribs
            if 'style' not in attribs:
                self._attribs['style'] = "display:none !important;"
                return
            # An existing style is in place
            style = self._attribs['style']
            style_list = style.split(';')
            # ensure no empty items
            style_list = [item for item in style_list if item]
            if "display:none !important" in style_list:
                # "display:none !important" already set
                return
            #  remove existing displays
            if "display:none" in style_list:
                # "display:none" already set, remove it so it can be replaced
                style_list.remove("display:none")
            if "display:block" in style_list:
                # "display:block" already set, remove it so it can be replaced
                style_list.remove("display:block")
            if "display:block !important" in style_list:
                # "display:block !important" already set, remove it so it can be replaced
                style_list.remove("display:block !important")
            # append display:none !important;
            if not style_list:
                self._attribs['style'] = "display:none !important;"
                return
            style_list.append("display:none !important;")
            self._attribs['style'] = ";".join(style_list)
        else:
            attribs = self._attribs
            if 'style' not in attribs:
                return
            # An existing style is in place
            style = self._attribs['style']
            style_list = style.split(';')
            style_list = [item for item in style_list if item]
            if "display:none !important" in style_list:
                style_list.remove("display:none !important")
            if "display:none" in style_list:
                style_list.remove("display:none")
            if style_list:
                self._attribs['style'] = ";".join(style_list)
            else:
                del self._attribs['style']

    def set_block(self):
        "Sets display:block !important into style"
        attribs = self._attribs
        if 'style' not in attribs:
            self._attribs['style'] = "display:block !important;"
            return
        # An existing style is in place
        style = self._attribs['style']
        style_list = style.split(';')
        # ensure no empty items
        style_list = [item for item in style_list if item]
        if "display:block !important" in style_list:
            # "display:block !important" already set
            return
        #  remove existing displays
        if "display:block" in style_list:
            # "display:block" already set, remove it so it can be replaced
            style_list.remove("display:block")
        if "display:none" in style_list:
            # "display:none" already set, remove it so it can be replaced
            style_list.remove("display:none")
        if "display:none !important" in style_list:
            # "display:none !important" already set, remove it so it can be replaced
            style_list.remove("display:none !important")
        # append display:block !important;
        if not style_list:
            self._attribs['style'] = "display:block !important;"
            return
        style_list.append("display:block !important;")
        self._attribs['style'] = ";".join(style_list)

    def get_attribs(self):
        return self._attribs.copy()

    def set_attribs(self, attribs):
        self._attribs = {}
        if attribs:
            for att,val in attribs.items():
                if val:
                    self._attribs[att]=val
                else:
                    self._attribs[att]=''

    def del_attribs(self):
        self._attribs = {}

    attribs = property(get_attribs, set_attribs, del_attribs)

    @property
    def attributes_string(self):
        "Returns a string showing the attributes"
        str_attribs = ''
        for att,val in self._attribs.items():
            if val:
                if val[0] == '{':
                    value = self._expand_label(str(val))
                else:
                    value = str(val)
                str_attribs += " {att!s}=\"{val!s}\"".format(att=att, val=html.escape(value))
            else:
                str_attribs += " {att!s}=\"\"".format(att=att)
        return str_attribs

    def _expand_label(self, valstring):
        "If valstring starts with a  label in the form {label}, substitute url here"
        if valstring[0] != '{':
            return valstring
        proj_ident = self.proj_ident
        if not proj_ident:
            return valstring
        end = valstring.find('}')
        if end == -1:
            return valstring
        label = valstring[1:end]
        if label:
            url = skiboot.get_url(label, proj_ident)
            if url:
                return quote(url, safe='/:?&=@')+valstring[end+1:]
        return valstring

    def has_attrib(self, name):
        "If this attrib exists, return True otherwise False"
        return name in self._attribs

    def get_attrib_value(self, name):
        """Get an attribute value, given its name, if it does not exist, return None"""
        if name not in self._attribs:
            return None
        return self._attribs[name]

    def del_one_attrib(self, attrib):
        "deletes an attribute"
        if attrib in self._attribs:
            del self._attribs[attrib]


    def update_attribs(self, attribs):
        "Updates attributes with the dictionary given in the attribs argument"
        for att,val in attribs.items():
            self._attribs[att]=val


    def make_get_url(self, page, url, get_fields={}, force_ident=False):
        "returns url, including the page ident_data for a get link"
        if not url:
            raise ServerError("URL is invalid")
        if not page.ident_data_string:
            return url
        getstring = ''
        for key, val in get_fields.items():
            if not val:
                continue
            getstring += "&" + quote_plus(key) + "=" + quote_plus(str(val))
            force_ident = True
        if not force_ident:
            return url
        return url + "?ident=" + quote_plus(page.ident_data_string) + getstring
 

    def show_error(self, message="", *args, **nargs):
        """If a message is given, sets self._error
        """
        # update attribute with  data-status="error"
        self.update_attribs({'data-status':'error'})
        if message:
            self._error = message
        self.show = True


    def __bool__(self):
        return True


    def string_indent(self, n=2):
        """Returns a text string to illustrate the element, indented by
           n spaces."""
        display_list = self.__str__().splitlines()
        if n == 2:
            return '  ' + '\n  '.join(display_list)
        sp = ' ' * n
        sp_n = '\n' + sp
        return sp + sp_n.join(display_list)

    def outline(self, proj_ident):
        """To be overwritten by child classes, returns a list defining the part used when creating JSON files representing the part"""
        return ['Text', str(self)]



class Part(ParentPart):
    """
    Defines a tagged part, such as a section, or paragraph with attributes and contents, the contents
    may be other Parts. The __str__ method returns the string of the tag.

    **instance attributes:**

    *attribs*
      A dictionary of attributes, key is attribute name, value is attribute value


    *parts*
      A list of sub parts

    *show*
      A True or false flag, normally True, which indicates the tag will be shown,
      if false the __str__ method returns an empty string

    *tag_name*
      such as 'p'

    *text*
      Text to place between the openning and closing tags - only inserted
      if the Part contains no sub parts.
    """

    # for formatting; any tag with these names will have its end tag put on
    # a new line
    tag_list = ["div", "head", "body", "nav", "section", "header", "footer", "form", "table", "tr", "svg"]

    def __init__(self, tag_name="div", attribs=None, text="", show=True, hide_if_empty=False, brief=''):
        ParentPart.__init__(self, tag_name=tag_name, attribs=attribs, show=show, brief=brief)
        # self.htmlescaped is True where a part string contents should be escaped
        # False where the string contents is further html
        self.htmlescaped=True
        # self.linebreaks is True where a part string contents newlines should be
        # changed to linebreaks
        self.linebreaks=True
        # self.parts is a list of sub parts
        if text:
            self.parts = [text]
        self.hide_if_empty = hide_if_empty


    def make_js(self, page, ident_list, environ, call_data, lang):
        "Auto generates the javascript for any containing widgets"
        if not self.parts: return
        if not self.show: return
        for part in self.parts:
            if hasattr(part, 'make_js'):
                part.make_js(page, ident_list, environ, call_data, lang)


    def get_text(self):
        "equivalent to getting self[0]"
        if self.parts:
            return self.parts[0]
        else:
            return ''

    def set_text(self, text):
        "equivalent to setting self[0]"
        if self.parts:
            self.parts[0]=text
        else:
            self.parts = [text]

    text = property(get_text, set_text, doc="Equivalent to self[0]; note may not be text despite this attributes name")


    def set_contained_values(self, values):
        """This may be called during a _build by a widget wishing to set values into contained widgets
           It is generally overwitten by widgets which are intended to be contained, and need to do something
           with information from their parent container"""
        for index, part in enumerate(self.parts):
            if hasattr(part, "set_contained_values"):
                part.set_contained_values(values)

        
    def update(self, page, ident_list, environ, call_data, lang, ident_string, placename='', embedded=('','',None)):
        "Override this to change run time information in this part and its sub parts"
        if placename:
            self.placename = placename
        if not self.ident_string:
            # item is a dynamic part, newly created within a widget
            self.embedded = embedded
            self.ident_string = ident_string
        try:
            if self._error:
                if isinstance(self._error, TextBlock):
                    self._error.update(page, ident_list, environ, call_data, lang, self.ident_string, self.placename, embedded)
                return
            for index, part in enumerate(self.parts):
                part_ident_string = self.ident_string + "-" + str(index)
                if hasattr(part, "update"):
                    part.update(page, ident_list, environ, call_data, lang, part_ident_string, self.placename, embedded)
        except ValidateError as e:
            if not e.ident_list:
                e.ident_list = ident_list
            raise


    def __setitem__(self, part_index, part_obj):
        "Sets sub part, part_index must be an integer"
        if not isinstance(part_index, int):
            raise TypeError('indeces must be integers')
        # so its an integer
        len_parts = len(self.parts)
        if part_index > len_parts:
            raise ValidateError("part index too large")
        if part_index == len_parts:
            self.parts.append(part_obj)
            return
        self.parts[part_index] = part_obj


    def append(self, part_obj):
        "Sets sub part"
        self.parts.append(part_obj)

    def __getitem__(self, part_index):
        "Returns the part object for the given index, part_index must be an integer"
        return self.parts[part_index]


    def __delitem__(self, part_index):
        "Deletes a part, part_index must be an integer"
        del self.parts[part_index]

    def clear(self):
        self.parts = []


    def set_location_value(self, location_list, value):
        "Set a value in the part at this position, location_list being either an integer, a tuple/list of indexes"
        if isinstance(location_list, str):
            try:
                location_list = int(location_list)
            except Exception:
                raise TypeError('Invalid location')
        if isinstance(location_list, int):
            self[location_list] = value
            return
        if not location_list:
            raise TypeError('Invalid location')
        if len(location_list) == 1:
            self[location_list[0]] = value
            return
        part = self
        for i in location_list[:-1]:
            part = part[i]
        part[location_list[-1]]= value

    def insert_location_value(self, location_list, value):
        "Inserts a value in the part at this position, location_list being either an integer, a tuple/list of indexes"
        if isinstance(location_list, str):
            try:
                location_list = int(location_list)
            except Exception:
                raise TypeError('Invalid location')
        if isinstance(location_list, int):
            self.insert(location_list, value)
            return
        if not location_list:
            raise TypeError('Invalid location')
        if len(location_list) == 1:
            self.insert(location_list[0], value)
            return
        part = self
        for i in location_list[:-1]:
            part = part[i]
        part.insert(location_list[-1], value)

    def get_location_value(self, location_list):
        "Get a value in the part at this position, location_list being either an integer, a tuple/list of indexes"
        if isinstance(location_list, str):
            try:
                location_list = int(location_list)
            except Exception:
                raise TypeError('Invalid location')
        if isinstance(location_list, int):
            return self.parts[location_list]
        if not location_list:
            raise TypeError('Empty location')
        part = self
        for i in location_list:
            part = part.parts[i]
        return part

    def del_location_value(self, location_list):
        "Deletes the value at this position, location_list being either an integer, a tuple/list of indexes"
        if isinstance(location_list, str):
            try:
                location_list = int(location_list)
            except Exception:
                raise TypeError('Invalid location')
        if isinstance(location_list, int):
            del self[location_list]
            return
        if not location_list:
            raise TypeError('Invalid location')
        if len(location_list) == 1:
            del self[location_list[0]]
            return
        part = self
        for i in location_list[:-1]:
            part = part.parts[i]
        del part[location_list[-1]]

    def __len__(self):
        return len(self.parts)

    def __iter__(self):
        return self.parts.__iter__()

    def __bool__(self):
        return bool(self.parts)

    def insert(self, index, part_obj):
        "Inserts a sub part at index position, index must be an integer"
        self.parts.insert(index, part_obj)


    def data(self):
        "Returns the part as a list"
        if not self.show:
            return []
        if self._error:
            if self.name:
                # insert data-status=error if this is a named widget.
                return ["<span data-status=\"error\">", expand_text(self._error), "</span>"]
            else:
                return ["<span>", expand_text(self._error), "</span>"]

        content = []
        for part in self.parts:
            if isinstance(part, str):
                if part:
                    if (self.tag_name == 'script') or not self.htmlescaped:
                        # append the string without applying html escape or inserting <br />
                        content.append(part)
                    elif (self.tag_name == 'pre') or (self.tag_name == 'textarea'):
                        # append the string, escape html but do not insert <br />
                        content.append(html.escape(part))
                    else:
                        # append the string, escape html and insert <br />  if self.linebreaks is True
                        content.append(expand_text(part, linebreaks=self.linebreaks))
            elif hasattr(part, 'data'):
                content.extend(part.data())
            else:
                str_part = str(part)
                if str_part:
                    content.append(str_part)
            if self.tag_name == "head":
                content.append("\n")
        if (not content) and self.hide_if_empty:
            return []
        # this creates some new lines around head and body tags
        if self.tag_name == "head":
            start_tag = "<{tag_name}{str_attribs}>\n".format(tag_name=self.tag_name, str_attribs=self.attributes_string)
            end_tag = "</{tag_name}>\n".format(tag_name=self.tag_name)
        elif self.tag_name == "body":
            start_tag = "<{tag_name}{str_attribs}>\n".format(tag_name=self.tag_name, str_attribs=self.attributes_string)
            end_tag = "\n</{tag_name}>".format(tag_name=self.tag_name)
        else:
            start_tag = "<{tag_name}{str_attribs}>".format(tag_name=self.tag_name, str_attribs=self.attributes_string)
            end_tag = "</{tag_name}>".format(tag_name=self.tag_name)
        partbytes = [start_tag]
        partbytes.extend(content)
        partbytes.append(end_tag)
        return partbytes


    def outline(self, proj_ident):
        "Creates a list of ['Part', dictionary]"
        part_dict = OrderedDict()
        part_dict["tag_name"] = self.tag_name
        if self.brief:
            part_dict["brief"] = self.brief
        part_dict["show"] = self.show
        part_dict["hide_if_empty"] = self.hide_if_empty
        if self.attribs:
            part_attribs = OrderedDict(sorted(self.attribs.items(), key=lambda t: t[0]))
            part_dict["attribs"] = make_dictionary(part_attribs, proj_ident)
        # now the Part contents in a list
        parts = []
        for item in self.parts:
            if hasattr(item, 'outline'):
                parts.append(item.outline(proj_ident))
            else:
                # must be a text string
                parts.append(['Text', str(item)])
        part_dict["parts"] = parts
        return ['Part', part_dict]


    def __repr__(self):
        return '<' + self.tag_name + '>'

    def __str__(self):
        "Returns a text string to illustrate the element"
        content = ""
        for part in self.parts:
            if isinstance(part, str):
                if part:
                    if (self.tag_name == 'pre') or (self.tag_name == 'script'):
                        content += part
                    elif self.tag_name == 'textarea':
                        content += html.escape(part)
                    else:
                        content += expand_text(part, linebreaks = self.linebreaks)
            elif isinstance(part, Part):
                content += part.string_indent()
            elif isinstance(part, ClosedPart):
                content += part.string_indent()
            else:
                content += str(part)
        start_tag = "\n<{tag_name}{str_attribs}>".format(tag_name=self.tag_name, str_attribs=self.attributes_string)
        if self.tag_name in self.tag_list:
            end_tag = "\n</{tag_name}>".format(tag_name=self.tag_name)
        else:
            end_tag = "</{tag_name}>".format(tag_name=self.tag_name)
        return start_tag+content+end_tag


class ClosedPart(ParentPart):
    """
    Defines a closed tagged part ending in />

    **instance attributes:**

    *attribs*
      A dictionary of attributes, key is attribute name, value is attribute value

    *show*
      A True or false flag, normally True, which indicates the tag will be shown,
      if false the __str__ method returns an empty string

    *tag_name*
      such as 'br'
    """

    def __init__(self, tag_name="br", attribs=None, show=True, brief=''):
        ParentPart.__init__(self, tag_name=tag_name, attribs=attribs, show=show, brief=brief)


    def make_js(self, page, ident_list, environ, call_data, lang):
        "Auto generates the javascript for this part, normally overwritten by child widgets"
        pass

    def update(self, page, ident_list, environ, call_data, lang, ident_string, placename='', embedded=('','',None)):
        "Override this to change run time information in this part"
        if placename:
            self.placename = placename
        if not self.ident_string:
            # item is a dynamic part, newly created within a widget
            self.embedded = embedded
            self.ident_string = ident_string
        if self._error:
            if isinstance(self._error, TextBlock):
                self._error.update(page, ident_list, environ, call_data, lang, ident_string, placename, embedded)

    def data(self):
        "Returns the part as a list"
        if not self.show:
            return []
        if self._error:
            if self.name:
                # insert data-status=error if this is a named widget.
                return ["<span data-status=\"error\">", expand_text(self._error), "</span>"]
            else:
                return ["<span>", expand_text(self._error), "</span>"]
        parttext = "<{tag_name}{str_attribs} />".format(tag_name=self.tag_name, str_attribs=self.attributes_string)
        return [parttext]


    def outline(self, proj_ident):
        "Creates a list of ['ClosedPart', dictionary]"
        part_dict = OrderedDict()
        part_dict["tag_name"] = self.tag_name
        if self.brief:
            part_dict["brief"] = self.brief
        part_dict["show"] = self.show
        if self.attribs:
            part_attribs = OrderedDict(sorted(self.attribs.items(), key=lambda t: t[0]))
            part_dict["attribs"] = make_dictionary(part_attribs, proj_ident)
        return ['ClosedPart', part_dict]


    def __repr__(self):
        return '<' + self.tag_name + ' />'


    def __str__(self):
        "Returns a text string to illustrate the element"
        return "\n<{tag_name}{str_attribs} />".format(tag_name=self.tag_name, str_attribs=self.attributes_string)


class Section(Part):
    "This is essentially a Part, but with extra widget dictionary and change attribute"

    def __init__(self, tag_name="div", attribs=None, text="", show=True, hide_if_empty=False, brief=''):
        "As a Part but also has self.widgets"
        Part.__init__(self, tag_name=tag_name, attribs=attribs, text=text, show=show, hide_if_empty=hide_if_empty, brief=brief)
        # self.widgets is a widget name -> widget dictionary
        self.widgets = {}
        # self.validator_scriptlinks is a list of validator module ski_name's
        # calculated when the section is saved, and used when the section is imported
        # into a page to add links to the validator js modules in the page head
        self._validator_scriptlinks = []
        # currently unused, though passed to parts in set_idents
        self.section_places = {}
        # the change uuid is regenerated
        # whenever the item is stored in the database
        self.change = uuid.uuid4().hex

    @property
    def validator_scriptlinks(self):
        return self._validator_scriptlinks

    def load_validator_scriptlinks(self):
        "Called when section saved to database"
        self._validator_scriptlinks = []
        val_modules = []
        for widget in self.widgets.values():
            for field in widget.fields.values():
                val_list = field.val_list
                for val in val_list:
                    mod_name = val.__class__.__module__.split(".")[-1]
                    if mod_name not in val_modules:
                        val_modules.append(mod_name)
        for mod_name in val_modules:
            link_label = 'ski_' + mod_name
            if link_label not in self._validator_scriptlinks:
                self._validator_scriptlinks.append(link_label)

    def update(self, page, ident_list, environ, call_data, lang, ident_string, placename='', embedded=('','',None)):
        "Override this to change run time information in this part and its sub parts"
        if placename:
            self.placename = placename
        if not self.ident_string:
            # item is a dynamic part, newly created within a widget
            self.embedded = embedded
            self.ident_string = ident_string
        try:
            if self._error:
                if isinstance(self._error, TextBlock):
                    self._error.update(page, ident_list, environ, call_data, lang, self.ident_string, self.placename, embedded)
                return
            for index, part in enumerate(self.parts):
                part_ident_string = self.ident_string + "-" + str(index)
                if hasattr(part, "update"):
                    part.update(page, ident_list, environ, call_data, lang, part_ident_string, self.placename, embedded)
        except ValidateError as e:
            if not e.ident_list:
                e.ident_list = ident_list
            raise

    def location_item(self, location):
        "Returns the part or widget at location"
        location_string, container_number, location_list = location
        if not location_string:
            return
        if container_number is None:
            # Not in a widget
            if location_list:
                return self.get_location_value(location_list)
            else:
                return self
        else:
            # location string must be a widget name
            if location_string not in self.widgets:
                return
            widget = self.widgets.get(location_string)
            if widget is None:
                return
            if widget.can_contain():
                if not location_list:
                    # return the container
                    return widget.container_part(container_number)
                else:
                    return widget.get_from_container(container_number, location_list)
        # part not found
        return

    def outline(self, proj_ident):
        "Creates a list of ['Section', dictionary]"
        part, part_dict = Part.outline(self, proj_ident)
        return ['Section', part_dict]


class SectionPlaceHolder(object):
    "Instance of this is added to a part, and acts as the placeholder for a section"

    def __init__(self, section_name, placename, multiplier=0, mtag="div", brief=''):
        self.brief = brief
        self.section_name = section_name
        self.placename = placename
        self.mtag = mtag
        self.ident_string = ''
        try:
            mult = int(multiplier)
        except Exception:
            mult = 0
        if mult<1:
            mult = 0
        self.multiplier=mult


    def get_section(self):
        if not self.ident_string:
            return
        proj_ident = self.ident_string.split('_')[0]
        proj = skiboot.getproject(proj_ident)
        if proj is None:
            return
        if self.section_name not in proj.list_section_names():
            return
        return proj.section(self.section_name)

    @property
    def pagepart(self):
        """pagepart is typically head, body or svg.
           ident_string normally starts project_pageident_head-0-... etc
        """
        if not self.ident_string:
            return ''
        page_list = self.ident_string.split('_', 2)
        part_list = page_list[2].split('-', 1)
        return part_list[0]

    @property
    def ident_list(self):
        if not self.ident_string:
            return []
        ident_list = self.ident_string.split('-')
        return [ int(p) for p in ident_list[1:] ]


    def set_idents(self, ident_string, widgets, section_places, embedded):
        "This is called when the page is added to a folder, adds this placeholder to the page section_places dictionary "
        self.ident_string = ident_string
        # add this placeholder to the page section_places dictionary
        section_places[self.placename] = self

    def set_unique_names(self, name_list):
        "Ensures this item has a unique name"
        number = 2
        new_name = self.placename
        while new_name in name_list:
            new_name = self.placename + str(number)
            number += 1
        self.placename = new_name 
        name_list.append(self.placename)

    def outline(self, proj_ident):
        """Creates a list of ['SectionPlaceHolder', dictionary]"""
        part_dict = OrderedDict()
        if self.brief:
            part_dict["brief"] = self.brief
        if self.section_name:
            part_dict["section_name"] = self.section_name
        if self.placename:
            part_dict["placename"] = self.placename
        part_dict["multiplier"] = self.multiplier
        part_dict["mtag"] = self.mtag
        return ['SectionPlaceHolder', part_dict]


    def __str__(self):
        return ''


class TextBlock(object):

    def __init__(self, textref='', project='', failmessage=None, escape=True, linebreaks=True, replace_strings=[], show = True):
        # Note project is either '' for a textblock defined in this project, or the proj_ident of a sub project
        self.project = project
        # set show to False if this textblock is not to be shown
        self.show = show
        self.textref = textref
        self._displayedtext = ""
        self._failmessage = failmessage
        self.linebreaks=linebreaks
        self.escape=escape
        # uses the % operator
        self.replace_strings=replace_strings
        self._show_failmessage = False
        # error message
        self._error_message = ''

    def __bool__(self):
        "True if a reference have been set"
        if self.textref:
            return True
        else:
            return False

    def show_error(self, message, message_ref):
        "Called by a widget containing this textblock if the widget wishes to display an error"
        self._error_message = message
        self.textref = message_ref
        # on error, do not place any replace strings
        self.replace_strings=[]

    def failmessage_set(self):
        "Returns True if failmessage has been set, False otherwise"
        return bool(self._failmessage)

    def get_failmessage(self):
        if self._failmessage:
            return self._failmessage
        if self.textref:
            return "failed to read textblocks with ref " + self.textref
        return "failed to read textblocks"

    def set_failmessage(self, message):
        self._failmessage = message

    failmessage = property(get_failmessage, set_failmessage)

    def update(self, page, ident_list, environ, call_data, lang, ident_string, placeholder, embedded):
        if self.project:
            proj_ident = self.project
        else:
            proj_ident = page.ident.proj
        self._show_failmessage = False
        if self._error_message: return
        if not self.textref:
            self._show_failmessage = True
            return
        proj = skiboot.getproject(proj_ident)
        if proj is None:
            self._show_failmessage = True
            return
        result = proj.textblocks.get_text(self.textref, lang)
        if result is None:
            self._show_failmessage = True
        else:
            self._displayedtext = result


    def exists(self, proj_ident=''):
        "Returns true if textref is available in the project"
        if not self.textref:
            return False
        if proj_ident:
            proj = skiboot.getproject(proj_ident)
        elif self.project:
            proj = skiboot.getproject(self.project)
        else:
            return False
        if proj is None:
            return False
        return proj.textblocks.textref_exists(self.textref)


    def outline(self, proj_ident):
        "Creates a list of ['TextBlock', dictionary]"
        part_dict = OrderedDict()
        part_dict["textref"] = self.textref
        if self.project and (self.project != proj_ident):
            part_dict["project"] = self.project
        if self.failmessage:
            part_dict["failmessage"] = self.failmessage
        part_dict["show"] = self.show
        part_dict["escape"] = self.escape
        part_dict["linebreaks"] = self.linebreaks
        if self.replace_strings:
            part_dict["replace_strings"] = make_list(self.replace_strings, proj_ident)
        return ['TextBlock', part_dict]

  
    def __str__(self):
        "Returns the text"
        if self._error_message:
            return expand_text(self._error_message, escape=self.escape, linebreaks=self.linebreaks)
        if self._show_failmessage:
            return expand_text(self.failmessage, escape=self.escape, linebreaks=self.linebreaks)
        return expand_text(self._displayedtext, escape=self.escape, linebreaks=self.linebreaks, replace_strings=self.replace_strings)


class HTMLSymbol(object):

    def __init__(self, text=''):
        self.text=text

    def __bool__(self):
        "True if text has been set"
        return bool(self.text)

    def outline(self, proj_ident):
        """Creates a list of ['HTMLSymbol', {'text':text}]"""
        part_dict = OrderedDict()
        part_dict["text"] = self.text
        return ['HTMLSymbol', part_dict]

    def __str__(self):
        "Returns the text"
        return self.text


class Comment(object):

    def __init__(self, text=''):
        self._text=text.replace("--", "  ")

    def get_text(self):
        return self._text

    def set_text(self, text):
        self._text=text.replace("--", "  ")

    text = property(get_text, set_text, doc="Text of the comment, not including <!-- and --> parts")

    def outline(self, proj_ident):
        """Creates a list of ['Comment', {'text':text}]"""
        part_dict = OrderedDict()
        part_dict["text"] = self._text
        return ['Comment', part_dict]

    def __bool__(self):
        "True if text has been set"
        return bool(self._text)

    def __str__(self):
        "Returns the text"
        return  "\n<!--" + self._text + "-->\n"







import re

# a search for anything none-alphanumeric and not an underscore or a dot
_TB = re.compile('[^\w\.]')


from skipole.skilift import item_info, editpage, editsection

from skipole import FailPage, ValidateError, GoTo, ServerError


def retrieve_textblockref(skicall):
    "Fills in the edit textblockref page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # a skilift.part_tuple is (project, pagenumber, page_part, section_name, name, location, part_type, brief)

    pagenumber = None
    section_name = None

    if 'part_tuple' in call_data:
        part_tuple = call_data['part_tuple']
        project = part_tuple.project
        location = part_tuple.location
        pagenumber = part_tuple.pagenumber
        section_name = part_tuple.section_name
    else:
        project = call_data['editedprojname']
        location = call_data['location']
        if 'page_number' in call_data:
            pagenumber = call_data['page_number']
        else:
            section_name = call_data['section_name']

    if (pagenumber is None) and (section_name is None):
        raise ValidateError("Page/section not identified")

    call_data['location'] = location

    tblock = None

    try:
        if pagenumber:
            call_data['page_number'] = pagenumber
            pchange = call_data['pchange']
            tblock = editpage.page_textblock(project, pagenumber, pchange, location)
        else:
            call_data['section_name'] = section_name
            schange = call_data['schange']
            tblock = editsection.section_textblock(project, section_name, schange, location)

        if tblock is None:
            raise FailPage("TextBlock not identified")
    except ServerError as e:
        raise FailPage(e.message)

    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Edit the TextBlock."

    # header done, now page contents

    page_data[("textblock_ref","input_text")]=tblock.textref
    page_data["tblock_project", "input_text"]=tblock.tblock_project
    page_data[("textblock_failed","input_text")]=tblock.failmessage

    page_data[("linebreaks","radio_values")]=['ON', 'OFF']
    page_data[("linebreaks","radio_text")]=['On', 'Off']
    if tblock.linebreaks:
        page_data[("linebreaks","radio_checked")] = 'ON'
    else:
        page_data[("linebreaks","radio_checked")] = 'OFF'

    page_data[("setescape","radio_values")]=['ON', 'OFF']
    page_data[("setescape","radio_text")]=['On', 'Off']
    if tblock.escape:
        page_data[("setescape","radio_checked")] = 'ON'
    else:
        page_data[("setescape","radio_checked")] = 'OFF'



def set_textblock(skicall):
    "Sets the textblock reference, fail message, linebreaks or escape"

    call_data = skicall.call_data
    page_data = skicall.page_data

    pagenumber = None
    section_name = None

    project = call_data['editedprojname']
    location = call_data['location']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
        pchange = call_data['pchange']
    else:
        section_name = call_data['section_name']
        schange = call_data['schange']

    if (pagenumber is None) and (section_name is None):
        raise ValidateError("Page/section not identified")

    try:
        if pagenumber:
            tblock = editpage.page_textblock(project, pagenumber, pchange, location)
        else:
            tblock = editsection.section_textblock(project, section_name, schange, location)

        message = "A new value to edit has not been found"

        if 'textblock_ref' in call_data:
            if _TB.search(call_data["textblock_ref"]):
                raise FailPage(message="Invalid reference")
            textref = call_data["textblock_ref"]
            message = 'New reference string set'
        else:
            textref = tblock.textref

        if 'tblock_project' in call_data:
            tblock_project = call_data['tblock_project']
            if tblock_project == project:
                message = "TextBlock sub project removed"
            elif tblock_project:
                message = 'New TextBlock sub project set'
            else:
                message = "TextBlock sub project removed"
        else:
            tblock_project = tblock.tblock_project

        if 'textblock_failed' in call_data:
            failmessage = call_data["textblock_failed"]
            message = 'New fail message set'
        else:
            failmessage = tblock.failmessage


        if 'linebreaks' in call_data:
            if call_data['linebreaks'] == 'ON':
                linebreaks = True
                message = 'Linebreaks set on'
            else:
                linebreaks = False
                message = 'Linebreaks set off'
        else:
            linebreaks = tblock.linebreaks


        if 'setescape' in call_data:
            if call_data['setescape'] == 'ON':
                escape = True
                message = 'HTML escape set on'
            else:
                escape = False
                message = 'HTML escape set off'
        else:
            escape = tblock.escape

        if pagenumber:
            call_data['pchange'] = editpage.edit_page_textblock(project, pagenumber, pchange, location, textref, tblock_project, failmessage, escape, linebreaks)
        else:
            call_data['schange'] = editsection.edit_section_textblock(project, section_name, schange, location, textref, tblock_project, failmessage, escape, linebreaks)

    except ServerError as e:
        raise FailPage(e.message)

    call_data['status'] = message


def create_insert(skicall):
    "Creates the textblock ref"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    # create the textblock
    if 'textblock_ref' not in call_data:
        raise FailPage(message="Invalid reference")
    if _TB.search(call_data["textblock_ref"]):
        raise FailPage(message="Invalid reference")
    textref = call_data['textblock_ref']

    if 'tblock_project' in call_data:
        tblock_project = call_data['tblock_project']
    else:
        tblock_project = ''

    if 'textblock_failed' in call_data:
        failmessage = call_data["textblock_failed"]
    else:
        failmessage = "TextBlock not found"

    if 'linebreaks' in call_data:
        if call_data['linebreaks'] == 'ON':
            linebreaks = True
        else:
            linebreaks = False
    else:
        linebreaks = True

    if 'setescape' in call_data:
        if call_data['setescape'] == 'ON':
            escape = True
        else:
            escape = False
    else:
        escape = True


    try:
        part_top, container, location_integers = call_data['location']

        if 'page_number' in call_data:
            pagenumber = call_data['page_number']
            page_info = item_info(project, pagenumber)
            if page_info is None:
                raise FailPage("Page to edit not identified")

            if (page_info.item_type != "TemplatePage") and (page_info.item_type != "SVG"):
                raise FailPage("Page not identified")

            # page to go back to
            target = None
            if part_top == 'head':
                target = "page_head"
            elif part_top == 'body':
                target = "page_body"
            elif part_top == 'svg':
                target = "page_svg"

            if container is not None:
                target = "back_to_container"

            if target is None:
                raise FailPage("Invalid location")

            call_data['pchange'], new_location = editpage.create_textblock_in_page(project,
                                                                                     pagenumber,
                                                                                     call_data['pchange'],
                                                                                     call_data['location'],
                                                                                     textref,
                                                                                     failmessage,
                                                                                     escape,
                                                                                     linebreaks,
                                                                                     tblock_project=tblock_project)

        elif 'section_name' in call_data:
            section_name = call_data['section_name']

            if container is not None:
                target = "back_to_container"
            else:
                target = "back_to_section"

            call_data['schange'], new_location = editsection.create_textblock_in_section(project,
                                                                                          section_name,
                                                                                          call_data['schange'],
                                                                                          call_data['location'],
                                                                                          textref,
                                                                                          failmessage,
                                                                                          escape,
                                                                                          linebreaks,
                                                                                          tblock_project=tblock_project)

        else:
            raise FailPage("Either a page or section must be specified")

    except ServerError as e:
        raise FailPage(e.message)

    call_data['location'] = new_location
    call_data['status'] = 'New TextBlock inserted'
    raise GoTo(target = target, clear_submitted=True)







import pkgutil, re, html, json

from .. import utils
from skipole import FailPage, ValidateError, GoTo, ServerError
from skipole.skilift import fromjson, part_info, part_contents, editsection, versions

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')

def retrieve_managepage(skicall):
    "this call is for the manage sections page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data
    utils.clear_call_data(call_data)

    project = call_data['editedprojname']

    page_data[("adminhead","page_head","large_text")] = "Manage Sections"

    # get current sections
    section_list = editsection.list_section_names(project)
    if not section_list:
        page_data[("tabledescription", "show")] = False
        page_data[("sectiontable", "show")] = False
        return

    # fill in the sections table

    #       contents: col 0 is the text to place in the first column,
    #                  col 1, 2, 3, 4 are the get field contents of links 1,2,3 and 4
    #                  col 5 - True if the first button and link is to be shown, False if not
    #                  col 6 - True if the second button and link is to be shown, False if not
    #                  col 7 - True if the third button and link is to be shown, False if not
    #                  col 8 - True if the fourth button and link is to be shown, False if not

    contents = []
    for section_name in section_list:
        contents.append([ section_name, section_name, section_name, '', '', True, True, False, False ])

    page_data[("sectiontable", "contents")] = contents



def retrieve_section_dom(skicall):
    "this call fills in the section dom table"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data, keeping section_name and schange
    utils.clear_call_data(call_data, ["section_name", "schange"])

    project = call_data['editedprojname']

    if "section_name" in call_data:
        section_name = call_data["section_name"]
    else:
        raise FailPage(message = "Section name missing")

    if not section_name:
        raise FailPage(message = "Section name missing")

    section_list = editsection.list_section_names(project)
    if section_name not in section_list:
        raise FailPage(message = "Section name invalid")

    call_data['schange'] = editsection.sectionchange(project, section_name)

    # fill in the section dom table

    # section location is a tuple of section_name, None for no container, () tuple of location integers
    section_location = (section_name, None, ())
    # get section_tuple from project, pagenumber, section_name, section_location
    section_tuple = part_info(project, None, section_name, section_location)

    # section_tuple is None if part not found, otherwise a namedtuple with items
    #  project, pagenumber, page_part, section_name, name, location, part_type, brief

    if section_tuple is None:
        raise FailPage("The section has not been recognised")

    try:
        domcontents, dragrows, droprows = _section_domcontents(project, section_name)
    except ServerError as e:
        raise FailPage(message=e.message)


    # widget editdom,domtable is populated with fields

    #    dragrows: A two element list for every row in the table, could be empty if no drag operation
    #              0 - True if draggable, False if not
    #              1 - If 0 is True, this is data sent with the call wnen a row is dropped
    #    droprows: A two element list for every row in the table, could be empty if no drop operation
    #              0 - True if droppable, False if not
    #              1 - text to send with the call when a row is dropped here
    #    dropident: ident or label of target, called when a drop occurs which returns a JSON page

    #    cols: A three element list for every column in the table, must be given with empty values if no links
    #              0 - target HTML page link ident of buttons in each column, if col1 not present or no javascript
    #              1 - target JSON page link ident of buttons in each column,
    #              2 - session storage key 'ski_part'

    #    contents: A list for every element in the table, should be row*col lists
    #               0 - text string, either text to display or button text
    #               1 - A 'style' string set on the td cell, if empty string, no style applied
    #               2 - Is button? If False only text will be shown, not a button, button class will not be applied
    #                       If True a link to link_ident/json_ident will be set with button_class applied to it
    #               3 - The get field value of the button link, empty string if no get field

    page_data['editdom', 'domtable', 'contents']  = domcontents
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows

    # for each column: html link, JSON link, storage key
    page_data['editdom', 'domtable', 'cols']  =  [    ['','',''],                                             # tag name, no link
                                                      ['','',''],                                             # brief, no link
                                                      ['no_javascript','move_up_in_section_dom',''],          # up arrow
                                                      ['no_javascript','move_up_right_in_section_dom',''],    # up right
                                                      ['no_javascript','move_down_in_section_dom',''],        # down
                                                      ['no_javascript','move_down_right_in_section_dom',''],  # down right
                                                      ['edit_section_dom','',''],                             # edit, html only
                                                      ['no_javascript','insert_in_section',''],               # insert/append
                                                      ['no_javascript',7580,''],                              # copy
                                                      ['no_javascript',7590,'ski_part'],                      # paste
                                                      ['no_javascript','cut_section_dom',''],                 # cut
                                                      ['no_javascript','delete_section_dom','']               # delete
                                                   ]

    page_data['editdom', 'domtable', 'dropident']  = 'move_in_section_dom'


def retrieve_section_contents(skicall):
    "this call is for the edit section contents page"
    # fill in section dom table
    retrieve_section_dom(skicall)
    # set page head text
    skicall.page_data[("adminhead","page_head","large_text")] = "Edit Section %s" % (skicall.call_data["section_name"],)


def submit_new_section(skicall):
    "Create new section"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    # get new section name
    if ("newsection", "section_name") not in call_data:
        raise FailPage(message = "Section name missing from call_data")
    section_name = call_data["newsection", "section_name"]
    if ('new_tag', 'input_text') in call_data:
        tag_name = call_data['new_tag', 'input_text']
    else:
        tag_name = 'div'
    if ('description', 'input_text') in call_data:
        brief = call_data['description', 'input_text']
    else:
        brief = "New Section"
    try:
        editsection.create_new_section(project, section_name, tag_name, brief)
    except (ValidateError, ServerError) as e:
        raise FailPage(message = e.message)
    utils.clear_call_data(call_data)
    call_data['status'] = 'Section %s created' % (section_name,)


def delete_section(skicall):
    "Deletes a section"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if "delete_section" not in call_data:
        raise FailPage(message = "Section name missing from call_data", widget="table_error")
    section_name = call_data["delete_section"]
    if not section_name:
        raise FailPage(message = "Section name missing", widget="table_error")
    section_list = editsection.list_section_names(project)
    if section_name not in section_list:
        raise FailPage(message = "Section name does not exists", widget="table_error")
    # deletes the section
    try:
        editsection.delete_section(project, section_name)
    except ServerError as e:
        raise FailPage(message = e.message, widget = "table_error")
    call_data['status'] = 'Section %s deleted' % (section_name,)


def downloadsection(skicall):
    "Gets section, and returns a json dictionary, this will be sent as an octet file to be downloaded"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if 'section_name' not in call_data:
        raise FailPage(message = "section missing")
    section_name = call_data["section_name"]
    project = call_data['editedprojname']

    parttext, part_dict = fromjson.section_outline(project, section_name)
    # set version and skipole as the first two items in the dictionary
    versions_tuple = versions(project)
    part_dict["skipole"] = versions_tuple.skipole
    part_dict.move_to_end('skipole', last=False)
    part_dict["version"] = versions_tuple.project
    part_dict.move_to_end('version', last=False)
    jsonstring = json.dumps(part_dict, indent=4, separators=(',', ':'))
    line_list = []
    n = 0
    for line in jsonstring.splitlines(True):
        binline = line.encode('utf-8')
        n += len(binline)
        line_list.append(binline)
    page_data['headers'] = [('content-type', 'application/octet-stream'), ('content-length', str(n))]
    return line_list



def newsectionpage(skicall):
    "Populate the page which creates a new section"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'new_section_name' not in call_data:
        raise FailPage(message = "new section name missing")
    section_name = call_data["new_section_name"]
    if not section_name:
        raise FailPage(message = "Section name missing")
    section_lower_name = section_name.lower()
    if (section_lower_name == 'body') or (section_lower_name == 'head') or (section_lower_name == 'svg'):
        raise FailPage(message="Unable to create the section, the name given is reserved")
    if _AN.search(section_name):
        raise FailPage(message="Invalid section name, alphanumeric and underscore only")
    if section_name[0] == '_':
        raise FailPage(message="Invalid section name, must not start with an underscore")
    if section_name.isdigit():
        raise FailPage(message="Unable to create the section, the name must include some letters")
    section_list = editsection.list_section_names(project)
    if section_name in section_list:
        raise FailPage(message = "Section name already exists")
    page_data[("adminhead","page_head","large_text")] = "New Section"
    page_data["description", "input_text"] = "New section %s" % (section_name,)
    # set hidden fields on the two forms with the submitted section name
    page_data["newsection", "section_name"] = section_name
    page_data["uploadsection","section_name"] = section_name


def file_new_section(skicall):
    "Create new section from uploaded file"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    # get new section name

    if ("uploadsection", "section_name") not in call_data:
        raise FailPage(message = "Section name missing from call_data")
    section_name = call_data["uploadsection", "section_name"]
    if not section_name:
        raise FailPage(message = "Section name missing")
    section_lower_name = section_name.lower()
    if (section_lower_name == 'body') or (section_lower_name == 'head') or (section_lower_name == 'svg'):
        raise FailPage(message="Unable to create the section, the name given is reserved")
    if _AN.search(section_name):
        raise FailPage(message="Invalid section name, alphanumeric and underscore only")
    if section_name[0] == '_':
        raise FailPage(message="Invalid section name, must not start with an underscore")
    if section_name.isdigit():
        raise FailPage(message="Unable to create the section, the name must include some letters")
    section_list = editsection.list_section_names(project)
    if section_name in section_list:
        raise FailPage(message = "Section name already exists", widget="new")

    # get file contents
    file_contents = call_data["uploadsection", "action"]
    json_string = file_contents.decode(encoding='utf-8')
    # create the section
    try:
        fromjson.create_section(project, section_name, json_string)
    except ServerError as e:
        raise FailPage(message = e.message)

    # new section created
    utils.clear_call_data(call_data)
    call_data['status'] = 'Section %s created' % (section_name,)


def edit_section_dom(skicall):
    "Called by domtable to edit an item in a section"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers
    # though string of integers may be missing if this is the section part itself

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers, so location_list[0] is the section name, location_integers is ()
        # edit the top section html part
        call_data['part_tuple'] = part_info(editedprojname, None, location_list[0], [location_list[0], None, ()])
        raise GoTo(target = 53007, clear_submitted=True)

    section_name = location_list[0]
    call_data['section_name'] = section_name
    call_data['schange'] = editsection.sectionchange(editedprojname, section_name)

    location_integers = [ int(i) for i in location_list[1:]]
    part_tuple = part_info(editedprojname, None, section_name, [section_name, None, location_integers])
    if part_tuple is None:
        raise FailPage("Item to edit has not been recognised")

    if part_tuple.name:
        # item to edit is a widget
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 54006, clear_submitted=True)
    if part_tuple.part_type == "Part":
        # edit the html part
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 53007, clear_submitted=True)
    if part_tuple.part_type == "ClosedPart":
        # edit the html closed part
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 53007, clear_submitted=True)
    if part_tuple.part_type == "HTMLSymbol":
        # edit the symbol
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 51107, clear_submitted=True)
    if part_tuple.part_type == "str":
        # edit the text
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 51017, clear_submitted=True)
    if part_tuple.part_type == "TextBlock":
        # edit the TextBlock
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 52017, clear_submitted=True)
    if part_tuple.part_type == "Comment":
        # edit the Comment
        call_data['part_tuple'] = part_tuple
        raise GoTo(target = 51207, clear_submitted=True)

    # note : a sectionplaceholder cannot appear in a section
    raise FailPage("Item to edit has not been recognised")



def copy_section(skicall):
    "Gets section part and return it in page_data['localStorage'] with key ski_part for browser session storage"
    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to copy missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers
    # though string of integers may be missing if this is the section part itself

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    section_name = location_list[0]
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers, so location_list[0] is the section name, location_integers is ()
        location_integers = ()
    else:
        location_integers = tuple( int(i) for i in location_list[1:] )
    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)

    # get a json string dump of the item outline, however change any Sections to Parts
    itempart, itemdict = fromjson.item_outline(editedprojname, None, section_name, location)
    if itempart == 'Section':
        jsonstring = json.dumps(['Part',itemdict], indent=0, separators=(',', ':'))
    else:
        jsonstring = json.dumps([itempart,itemdict], indent=0, separators=(',', ':'))
    page_data['localStorage'] = {'ski_part':jsonstring}
    call_data['status'] = 'Item copied, and can now be pasted.'


def paste_section(skicall):
    "Gets submitted json string and inserts it"
    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "position to paste missing")
    if ('editdom', 'domtable', 'cols') not in call_data:
        raise FailPage(message = "item to paste missing")
    json_string = call_data['editdom', 'domtable', 'cols']

    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers
    # though string of integers may be missing if this is the section part itself

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    section_name = location_list[0]
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers, so location_list[0] is the section name, location_integers is ()
        location_integers = ()
    else:
        location_integers = tuple( int(i) for i in location_list[1:] )
    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    call_data['schange'] = editsection.create_item_in_section(editedprojname, section_name, call_data['schange'], location, json_string)
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def cut_section_dom(skicall):
    "Called by domtable to cut an item in a section, and copy for later pasting"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers
        raise FailPage("Item to remove has not been recognised")

    location_integers = tuple( int(i) for i in location_list[1:] )
    section_name = location_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_tuple = part_info(editedprojname, None, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to remove has not been recognised")

    # prior to deleting, take a copy
    # get a json string dump of the item outline, however change any Sections to Parts
    itempart, itemdict = fromjson.item_outline(editedprojname, None, section_name, location)
    if itempart == 'Section':
        jsonstring = json.dumps(['Part',itemdict], indent=0, separators=(',', ':'))
    else:
        jsonstring = json.dumps([itempart,itemdict], indent=0, separators=(',', ':'))
    page_data['localStorage'] = {'ski_part':jsonstring}

    # remove the item
    try:
        call_data['schange'] = editsection.del_location(editedprojname, section_name, call_data['schange'], location)
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents

    # once item is deleted, no info on the item should be
    # left in call_data - this may not be required in future
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    call_data['status'] = 'Item copied and then deleted. Use paste to recover or move it.'


def delete_section_dom(skicall):
    "Called by domtable to delete an item in a section"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers
        raise FailPage("Item to remove has not been recognised")

    location_integers = tuple( int(i) for i in location_list[1:] )
    section_name = location_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_tuple = part_info(editedprojname, None, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to remove has not been recognised")

    # remove the item
    try:
        call_data['schange'] = editsection.del_location(editedprojname, section_name, call_data['schange'], location)
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents

    # once item is deleted, no info on the item should be
    # left in call_data - this may not be required in future
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    call_data['status'] = 'Item deleted.'



def move_up_in_section_dom(skicall):
    "Called by domtable to move an item in a section up"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers
        return
    else:
        location_integers = tuple( int(i) for i in location_list[1:] )
    section_name = location_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_tuple = part_info(editedprojname, None, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to move has not been recognised")

    if (len(location_integers) == 1) and (location_integers[0] == 0):
        # at top, cannot be moved
        raise FailPage("Cannot be moved up")

    if location_integers[-1] == 0:
        # move up to next level
        new_location_integers = location_integers[:-1]
    else:
        # swap parts on same level
        new_location_integers = list(location_integers[:-1])
        new_location_integers.append(location_integers[-1] - 1)

    # after a move, location is wrong, so remove from call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_top' in call_data:
        del call_data['part_top']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    # move the item
    try:
        call_data['schange'] = editsection.move_location(editedprojname, section_name, call_data['schange'], location, (section_name, None, new_location_integers))
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_up_right_in_section_dom(skicall):
    "Called by domtable to move an item in a section up and to the right"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers, so location_list[0] is the section name
        return
    else:
        location_integers = tuple( int(i) for i in location_list[1:] )
    section_name = location_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_tuple = part_info(editedprojname, None, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to move has not been recognised")

    if location_integers[-1] == 0:
        # at top of a part, cannot be moved
        raise FailPage("Cannot be moved up")
    new_parent_integers = list(location_integers[:-1])
    new_parent_integers.append(location_integers[-1] - 1)
    new_parent_location = (section_name, None, new_parent_integers)

    new_parent_tuple = part_info(editedprojname, None, section_name, new_parent_location)

    if new_parent_tuple is None:
        raise FailPage("Cannot be moved up")
    if new_parent_tuple.part_type != "Part":
        raise FailPage("Cannot be moved up")

    items_in_new_parent = len(part_contents(editedprojname, None, section_name, new_parent_location))

    new_location_integers =  tuple(new_parent_integers + [items_in_new_parent])

    # after a move, location is wrong, so remove from call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_top' in call_data:
        del call_data['part_top']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    # move the item
    try:
        call_data['schange'] = editsection.move_location(editedprojname, section_name, call_data['schange'], location, (section_name, None, new_location_integers))
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_down_in_section_dom(skicall):
    "Called by domtable to move an item in a section down"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers, the section top cannot be moved
        return
    else:
        location_integers = tuple( int(i) for i in location_list[1:] )
    section_name = location_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_tuple = part_info(editedprojname, None, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to move has not been recognised")

    if len(location_integers) == 1:
        # Just at immediate level below top
        parent_location = (section_name, None, ())
        items_in_parent = len(part_contents(editedprojname, None, section_name, parent_location))
        if location_integers[0] == (items_in_parent-1):
            # At end, cannot be moved
            raise FailPage("Cannot be moved down")
        new_location_integers = (location_integers[0]+2,)
    else:
        parent_integers = tuple(location_integers[:-1])
        parent_location = (section_name, None, parent_integers)
        items_in_parent = len(part_contents(editedprojname, None, section_name, parent_location))
        if location_integers[-1] == (items_in_parent-1):
            # At end of a part, so move up a level
            new_location_integers = list(parent_integers[:-1])
            new_location_integers.append(parent_integers[-1] + 1)
        else:
            # just insert into current level
            new_location_integers = list(parent_integers)
            new_location_integers.append(location_integers[-1] + 2)

    # after a move, location is wrong, so remove from call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_top' in call_data:
        del call_data['part_top']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    # move the item
    try:
        call_data['schange'] = editsection.move_location(editedprojname, section_name, call_data['schange'], location, (section_name, None, new_location_integers))
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def move_down_right_in_section_dom(skicall):
    "Called by domtable to move an item in a section down and to the right"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'contents') not in call_data:
        raise FailPage(message = "item to edit missing")
    editedprojname = call_data['editedprojname']
    part = call_data['editdom', 'domtable', 'contents']

    # so part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_list = part.split('-')
    # first item should be a string, rest integers
    if len(location_list) == 1:
        # no location integers, the section top cannot be moved
        return
    else:
        location_integers = tuple( int(i) for i in location_list[1:] )
    section_name = location_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location = (section_name, None, location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_tuple = part_info(editedprojname, None, section_name, location)
    if part_tuple is None:
        raise FailPage("Item to move has not been recognised")

    if len(location_integers) == 1:
        parent_location = (section_name, None, ())
    else:
        parent_integers = list(location_integers[:-1])
        parent_location = (section_name, None, parent_integers)
    items_in_parent = len(part_contents(editedprojname, None, section_name, parent_location))
    if location_integers[-1] == (items_in_parent-1):
        # At end of a block, cannot be moved
        raise FailPage("Cannot be moved down")
    new_parent_integers = list(location_integers[:-1])
    new_parent_integers.append(location_integers[-1] + 1)
    new_parent_location = (section_name, None, new_parent_integers)
    new_parent_tuple = part_info(editedprojname, None, section_name, new_parent_location)

    if new_parent_tuple is None:
        raise FailPage("Cannot be moved down")
    if not (new_parent_tuple.part_type == 'Part' or new_parent_tuple.part_type == 'Section'):
        raise FailPage("Cannot be moved down")

    new_location_integers = tuple(new_parent_integers+[0])

    # after a move, location is wrong, so remove from call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_top' in call_data:
        del call_data['part_top']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    # move the item
    try:
        call_data['schange'] = editsection.move_location(editedprojname, section_name, call_data['schange'], location, (section_name, None, new_location_integers))
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents



def move_in_section_dom(skicall):
    "Called by domtable to move an item in a section after a drag and drop"

    call_data = skicall.call_data
    page_data = skicall.page_data

    if ('editdom', 'domtable', 'dragrows') not in call_data:
        raise FailPage(message = "item to drop missing")
    editedprojname = call_data['editedprojname']
    part_to_move = call_data['editdom', 'domtable', 'dragrows']

    # so part_to_move is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    location_to_move_list = part_to_move.split('-')
    # first item should be a string, rest integers
    if len(location_to_move_list) == 1:
        # no location integers, the section top cannot be moved
        return
    else:
        location_to_move_integers = tuple( int(i) for i in location_to_move_list[1:] )
    section_name = location_to_move_list[0]

    # location is a tuple of section_name, None for no container, tuple of location integers
    location_to_move = (section_name, None, location_to_move_integers)
    # get part_tuple from project, pagenumber, section_name, location
    part_to_move_tuple = part_info(editedprojname, None, section_name, location_to_move)
    if part_to_move_tuple is None:
        raise FailPage("Item to move has not been recognised")


    # new location

    target_part = call_data['editdom', 'domtable', 'droprows']

    # so target_part is section name with location string of integers

    # create location which is a tuple or list consisting of three items:
    # a string of section name
    # a container integer, in this case always None
    # a tuple or list of location integers
    target_location_list = target_part.split('-')
    # first item should be a string, rest integers
    if len(target_location_list) == 1:
        # no location integers
        target_location_integers = ()
    else:
        target_location_integers = tuple( int(i) for i in target_location_list[1:] )

    if section_name != target_location_list[0]:
        raise FailPage("Target location has not been recognised")

    # location is a tuple of section_name, None for no container, tuple of location integers
    target_location = (section_name, None, target_location_integers)
    # get part_tuple from project, pagenumber, section_name, location
    target_part_tuple = part_info(editedprojname, None, section_name, target_location)
    if target_part_tuple is None:
        raise FailPage("Target has not been recognised")

    if (target_part_tuple.part_type == "Part") or (target_part_tuple.part_type == "Section"):
        # insert
        if target_location_integers:
            new_location_integers = list(target_location_integers)
            new_location_integers.append(0)
        else:
            new_location_integers = [0]
    else:
        # append
        new_location_integers = list(target_location_integers)
        new_location_integers[-1] = new_location_integers[-1] + 1

    # after a move, location is wrong, so remove from call_data
    if 'location' in call_data:
        del call_data['location']
    if 'part' in call_data:
        del call_data['part']
    if 'part_top' in call_data:
        del call_data['part_top']
    if 'part_loc' in call_data:
        del call_data['part_loc']

    # move the item
    try:
        call_data['schange'] = editsection.move_location(editedprojname, section_name, call_data['schange'], location_to_move, (section_name, None, new_location_integers))
    except ServerError as e:
        raise FailPage(message = e.message)

    # and re-draw the table
    domcontents, dragrows, droprows = _section_domcontents(editedprojname, section_name)
    page_data['editdom', 'domtable', 'dragrows']  = dragrows
    page_data['editdom', 'domtable', 'droprows']  = droprows
    page_data['editdom', 'domtable', 'contents']  = domcontents


def _section_domcontents(project, section_name):
    "Return the info for domtable contents"
    section_location = (section_name, None, ())
    parttext,partdict = fromjson.item_outline(project, None, section_name, section_location)
    # create first row of the table
    if "attribs" in partdict:
        section_tag = '&lt;' + partdict['tag_name'] + ' ... &gt;'
    else:
        section_tag = '&lt;' + partdict['tag_name'] + '&gt;'
    section_brief = html.escape(partdict['brief'])
    if len( section_brief)>40:
        section_brief =  section_brief[:35] + '...'
    if not section_brief:
         section_brief = '-'
    domcontents = [
                   [section_tag, '', False, '' ],
                   [section_brief, '', False, '' ],
                   ['', '', False, '' ],                                             # no up arrow for top line
                   ['', '', False, '' ],                                             # no up_right arrow for top line
                   ['', '', False, '' ],                                             # no down arrow for top line
                   ['', '', False, '' ],                                             # no down_right arrow for top line
                   ['Edit',  'width : 1%;', True, section_name],                     # edit
                   ['Insert','width : 1%;text-align: center;', True, section_name],  # insert
                   ['Copy','width : 1%;text-align: center;', True, section_name],  # copy image for top line
                   ['Paste','width : 1%;text-align: center;', True, section_name],  # paste image for top line
                   ['', '', False, '' ],                                             # no cut image for top line
                   ['', '', False, '' ]                                              # no delete image for top line
                  ]
    # add further items to domcontents
    part_string_list = []

    if 'parts' not in partdict:
        rows = 1
    else:
        rows = _domtree(partdict, section_name, domcontents, part_string_list)
    # for every row in the table
    dragrows = [ [ False, '']]
    droprows = [ [ True, section_name ]]

    # for each row (minus 1 as the first row is done)
    if rows > 1:
        for row in range(0, rows-1):
            dragrows.append( [ True, part_string_list[row]] )
            droprows.append( [ True, part_string_list[row]] )
    
    return domcontents, dragrows, droprows



def _domtree(partdict, part_loc, contents, part_string_list, rows=1, indent=1):
    "Creates the contents of the domtable"

    # note: if in a container
    # part_loc = widget_name + '-' + container_number
    # otherwise part_loc = body, head, svg, section_name


    indent += 1
    padding = "padding-left : %sem;" % (indent,)
    u_r_flag = False
    last_row_at_this_level = 0

    parts = partdict['parts']

    # parts is a list of items
    last_index = len(parts)-1

    #Text   #characters..      #up  #up_right  #down  #down_right   #edit   #insert  #copy  #paste  #cut #delete

    for index, part in enumerate(parts):
        part_location_string = part_loc + '-' + str(index)
        part_string_list.append(part_location_string)
        rows += 1
        part_type, part_dict = part
        # the row text
        if part_type == 'Widget' or part_type == 'ClosedWidget':
            part_name = 'Widget ' + part_dict['name']
            if len(part_name)>40:
                part_name = part_name[:35] + '...'
            contents.append([part_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        elif part_type == 'TextBlock':
            contents.append(['TextBlock', padding, False, ''])
            part_ref = part_dict['textref']
            if len(part_ref)>40:
                part_ref = part_ref[:35] + '...'
            if not part_ref:
                part_ref = '-'
            contents.append([part_ref, '', False, ''])
        elif part_type == 'SectionPlaceHolder':
            section_name = part_dict['placename']
            if section_name:
                section_name = "Section " + section_name
            else:
                section_name = "Section -None-"
            if len(section_name)>40:
                section_name = section_name[:35] + '...'
            contents.append([section_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        elif part_type == 'Text':
            contents.append(['Text', padding, False, ''])
            # in this case part_dict is the text string rather than a dictionary
            if len(part_dict)<40:
                part_str = html.escape(part_dict)
            else:
                part_str = html.escape(part_dict[:35] + '...')
            if not part_str:
                part_str = '-'
            contents.append([part_str, '', False, ''])
        elif part_type == 'HTMLSymbol':
            contents.append(['Symbol', padding, False, ''])
            part_text = part_dict['text']
            if len(part_text)<40:
                part_str = html.escape(part_text)
            else:
                part_str = html.escape(part_text[:35] + '...')
            if not part_str:
                part_str = '-'
            contents.append([part_str, '', False, ''])
        elif part_type == 'Comment':
            contents.append(['Comment', padding, False, ''])
            part_text = part_dict['text']
            if len(part_text)<33:
                part_str =  "&lt;!--" + part_text + '--&gt;'
            else:
                part_str = "&lt;!--" + part_text[:31] + '...'
            if not part_str:
                part_str = '&lt;!----&gt;'
            contents.append([part_str, '', False, ''])
        elif part_type == 'ClosedPart':
            if 'attribs' in part_dict:
                tag_name = "&lt;%s ... /&gt;" % part_dict['tag_name']
            else:
                tag_name = "&lt;%s /&gt;" % part_dict['tag_name']
            contents.append([tag_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        elif part_type == 'Part':
            if 'attribs' in part_dict:
                tag_name = "&lt;%s ... &gt;" % part_dict['tag_name']
            else:
                tag_name = "&lt;%s&gt;" % part_dict['tag_name']
            contents.append([tag_name, padding, False, ''])
            part_brief = html.escape(part_dict.get('brief',''))
            if len(part_brief)>40:
                part_brief = part_brief[:35] + '...'
            if not part_brief:
                part_brief = '-'
            contents.append([part_brief, '', False, ''])
        else:
            contents.append(['UNKNOWN', padding, False, ''])
            contents.append(['ERROR', '', False, ''])

        # UP ARROW
        if rows == 2:
            # second line in table cannot move upwards
            contents.append(['', '', False, '' ])
        else:
            contents.append(['&uarr;', 'width : 1%;', True, part_location_string])

        # UP RIGHT ARROW
        if u_r_flag:
            contents.append(['&nearr;', 'width : 1%;', True, part_location_string])
        else:
            contents.append(['', '', False, '' ])

        # DOWN ARROW
        if (indent == 2) and (index == last_index):
            # the last line at this top indent has been added, no down arrow
            contents.append(['', '', False, '' ])
        else:
            contents.append(['&darr;', 'width : 1%;', True, part_location_string])

        # DOWN RIGHT ARROW
        # set to empty, when next line is created if down-right not applicable
        contents.append(['', '', False, '' ])

        # EDIT
        contents.append(['Edit', 'width : 1%;', True, part_location_string])

        # INSERT or APPEND
        if part_type == 'Part':
            contents.append(['Insert', 'width : 1%;text-align: center;', True, part_location_string])
        else:
            contents.append(['Append', 'width : 1%;text-align: center;', True, part_location_string])

        # COPY
        contents.append(['Copy', 'width : 1%;', True, part_location_string])

        # PASTE
        contents.append(['Paste', 'width : 1%;', True, part_location_string])

        # CUT
        contents.append(['Cut', 'width : 1%;', True, part_location_string])

        # DELETE
        contents.append(['Delete', 'width : 1%;', True, part_location_string])

        u_r_flag = False
        if part_type == 'Part':
            if last_row_at_this_level and (part_dict['tag_name'] != 'script') and (part_dict['tag_name'] != 'pre'):
                # add down right arrow in previous row at this level, get loc_string from adjacent edit cell
                editcell = contents[last_row_at_this_level *12-6]
                loc_string = editcell[3]
                contents[last_row_at_this_level *12-7] = ['&searr;', 'width : 1%;', True, loc_string]
            last_row_at_this_level = rows
            rows = _domtree(part_dict, part_location_string, contents, part_string_list, rows, indent)
            # set u_r_flag for next item below this one
            if  (part_dict['tag_name'] != 'script') and (part_dict['tag_name'] != 'pre'):
                u_r_flag = True
        else:
            last_row_at_this_level =rows

    return rows







"""A WSGI application generator

The package makes the following available:

version - a version string of the form a.b.c

WSGIApplication - an instance of this class is a callable WSGI application (see below)

set_debug(mode) - a function to turn on debugging (if mode is True), or off (if mode is False)

use_submit_list - is available to optionally wrap the user defined submit_data function
                  Enables a responder 'submit list' to define package,module,function to
                  be called as the responder's submit_data function, where package,module
                  is relative to the users code.


Exceptions
----------

These are also provided, and can be raised within the users code:

ValidateError - returns the project validation error page

ServerError - returns the project server error page

GoTo - diverts the call to another page

FailPage - diverts the call to the calling Responder's 'Fail page'


WSGIApplication
---------------

An instance of this class should be created, and is a callable WSGI application.

The WSGIApplication has the following arguments which should be provided to create
an instance:

project - the project name

projectfiles - the directory containing your projects

proj_data - an optional dictionary, passed as an attribute of skicall,
            which is an object passed to your functions during the progress of a call

start_call - a function you should create, called at the start of a call

submit_data - a function you should create, called by responders

end_call - a function you should create, called at the end of the call, prior to returning the page

url - path where this project will be served, typically '/'

The WSGIApplication class has the methods:

__call__(self, environ, start_response) - which sets the instance as callable

set_accesstextblocks(self, accesstextblocks) - Set an instance of a class which reads and writes TextBlocks.

Normally not used, as a default class (defined in the skipole.textblocks module) is used. If required, your
own object could be created and set into the WSGIApplication with this method. If you do this, your object
should support the same methods of the default class.

add_project(self, proj, url) - adds other projects to the 'root' project.

Where proj is another instance of a WSGIApplication and will be served at the path
given by argument url. At the minimum, your project should add the 'skis' sub project
which provides needed javascript files. For example:

my_application.add_project(skis_application, url='/lib')

Which causes the skis project to be served at /lib.

Sub-projects (such as skis above) cannot have further sub projects added - all
sub-projects can only be added to the 'root' project.


python3 -m skipole
------------------

skipole can also be run from the command line with the python -m option

Usage is

python3 -m skipole mynewproj /path/to/projectfiles

Which creates a directory /path/to/projectfiles
containing three sub directories:

mynewproj - a subdirectory containing a new project
skis - a project serving needed javascript files, necessary for all projects
skiadmin - a project used to help develop 'mynewproj'

You should replace 'mynewproj' with your preferred name for a new project.
This is an optional argument. If not given then only the skis and skiadmin
projects will be created.

The path "/path/to/projectfiles" must be given, and is the path to a directory
where you will develop your project. Multiple projects can be created in one
'projectfiles' directory, or you could have multiple such directories holding
different projects.

If the directory already exists, any existing skiadmin and skis projects will
be overwritten, this facility is used to upgrade skis and skiadmin if a new
version of skipole is installed.

If mynewproj already exists in the directory, it will not be changed.

You should then inspect the file

/path/to/projectfiles/mynewproj.py

where your code will be developed.

"""

import sys, traceback, inspect

from functools import wraps
from importlib import import_module

from .ski import skiboot

from .ski.project_class_definition import SkipoleProject
from .ski.excepts import ValidateError, ServerError, GoTo, FailPage

version = skiboot.version()


__all__ = ['WSGIApplication', 'ValidateError', 'ServerError', 'GoTo', 'FailPage', 'set_debug', 'use_submit_list', 'version']


class WSGIApplication(object):
    """The WSGIApplication - an instance being a callable WSGI application"""

    def __init__(self, project, projectfiles, proj_data={}, start_call=None, submit_data=None, end_call=None, url="/"):
        """An instance of this class is a callable WSGI application.

Arguments are:

project - the project name

projectfiles - the directory containing your projects

proj_data - an optional dictionary, passed as an attribute of skicall,
            which is an object passed to your functions during the progress of a call

start_call - a function you should create, called at the start of a call

submit_data - a function you should create, called by responders

end_call - a function you should create, called at the end of the call, prior to returning the page

url - path where this project will be served, typically '/'
"""
        self._skipoleproject = SkipoleProject(project, projectfiles, proj_data, start_call, submit_data, end_call, url)


    @property
    def project(self):
        """The project name"""
        return self._skipoleproject.proj_ident

    @property
    def projectfiles(self):
        """The projectfiles path"""
        return self._skipoleproject.projectfiles

    @property
    def proj_data(self):
        """The proj_data dictionary"""
        return self._skipoleproject.proj_data

    @property
    def start_call(self):
        """The start_call function"""
        return self._skipoleproject.start_call

    @property
    def submit_data(self):
        """The submit_data function"""
        return self._skipoleproject.submit_data

    @property
    def end_call(self):
        """The end_call function"""
        return self._skipoleproject.end_call

    @property
    def url(self):
        """This project path"""
        return self._skipoleproject.url

    @property
    def brief(self):
        """This project brief description"""
        return self._skipoleproject.brief

    @property
    def version(self):
        """This project version string"""
        return self._skipoleproject.version

    def __call__(self, environ, start_response):
        "The instance is callable"
        return self._skipoleproject(environ, start_response)

    def add_project(self, proj, url=None):
        """Add a sub project to this root project, returns the sub project url
           proj is the sub project WSGIApplication object."""
        return self._skipoleproject.add_project(proj._skipoleproject, url)

    def set_accesstextblocks(self, accesstextblocks):
        """Set an instance of a class which reads and writes TextBlocks. The default class is defined in the skipole.textblocks module,
           which simply stores TextBlocks in memory after reading them from a JSON file, and is not suitable for the dynamic creation
           of TextBlocks. It is sufficient for the single user of Skiadmin, but is essentially read only when serving your pages.
           If required, your own object could be created which should support the methods of the default class, and could read and write
           from your own database. This opens the possibility of dynamic creation and alteration of TextBlocks."""
        self._skipoleproject.textblocks = accesstextblocks


def set_debug(mode):
    "If mode is True, this sets increased debug error messages to be displayed"
    skiboot.set_debug(mode)



# This 'use_submit_list' is available to wrap the submit_data function if required


def use_submit_list(submit_data):
    "Used to decorate submit_data to enable submit_list to define package,module,function"
    @wraps(submit_data)
    def submit_function(skicall):
        "This function replaces submit_data, if skicall.submit_list has two or more elements"
        if not skicall.submit_list:
            # do nothing, simply return the original submit_data
            return submit_data(skicall)
        if len(skicall.submit_list) < 2:
            # do nothing, simply return the original submit_data
            return submit_data(skicall)
        # get the module where submit_data is defined
        sdmodule = inspect.getmodule(submit_data)
        if sdmodule.__name__ == "__main__":
            # absolute path
            submitpath = ".".join(skicall.submit_list[:-1])
            sdpackage = None
        else:
            # relative path
            submitpath = "." + ".".join(skicall.submit_list[:-1])
            sdpackage = sdmodule.__name__
        try:
            submitmodule = import_module(submitpath, sdpackage)
        except Exception as e:
            raise ServerError("Unable to import project module defined as %s in submit list" % (submitpath,)) from e
        # now obtain and run the specified function
        try:
            submitfunc = getattr(submitmodule, skicall.submit_list[-1])
        except Exception as e:
            raise ServerError("submit_list package %s found, but the required function %s is not recognised" % (submitpath, skicall.submit_list[-1])) from e
        return submitfunc(skicall)
    return submit_function










"""Contains commonly used paragraphs"""

from string import Template
import json

from .. import skiboot
from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class TagBlock(Widget):

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'tag':FieldArg("text", 'div'),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'drag':FieldArg("text", '', valdt=True, jsonset=True),
                        'drop':FieldArg("text", '', valdt=True, jsonset=True),
                        'dropident':FieldArg("url", "")}

    _container = ((),)

    def __init__(self, name=None, brief='', **field_args):
        """Acts as a widget, containing other widgets, so show, class and hide can be set
           drag - if given this is text data sent with a JSON call when this item is dropped
                  if nothing set here, this item will not be set as draggable.
           drop - if given this is text data sent with a JSON call when another item is dropped here
                  if nothing set here, this item will not be set as droppable.
           dropident - ident or label of target which returns a JSON page, called when a drop occurs here

        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # container is the widget itself
        self[0] =  ""
        self._dropurl = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self.tag_name = self.get_field_value('tag')
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        # dropurl
        self._dropurl = skiboot.get_url(self.get_field_value("dropident"), proj_ident=page.proj_ident)
        # drag
        drag = self.get_field_value("drag")
        if drag:
            self.update_attribs(
{"draggable":"true",
"ondragstart":"SKIPOLE.widgets['{ident}'].dragstartfunc(event, '{data}')".format(ident = self.get_id(),
                                                                                 data = drag)})
        # drop
        drop = self.get_field_value("drop")
        if drop:
            self.update_attribs(
{"ondrop":"SKIPOLE.widgets['{ident}'].dropfunc(event, '{data}')".format(ident = self.get_id(), data = drop),
"ondragover":"SKIPOLE.widgets['{ident}'].allowdropfunc(event)".format(ident = self.get_id())})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets drop and drag"""
        if self._dropurl:
            return self._make_fieldvalues(dropurl=self._dropurl)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class, and with div tag, or any other specified -->
       <!-- attribute style=display:none if hide is True, with drag and drop events if enabled -->
  <!-- further html and widgets can be contained here -->
</div>"""


class DivStyleDiv(Widget):

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'inner_tag':FieldArg("text", 'div'),
                        'style':FieldArg("cssstyle", ''),
                        'set_html':FieldArg("text", "", jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        """A div, containing a div which can have a style set, containing unescaped html
            inner_tag - the tag of the inside element
           style - the style of the inside element"""
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='div')
        self[0].htmlescaped = False
        self[0].linebreaks=False
        self[0][0] =  ""  # where html is to be set

    def _build(self, page, ident_list, environ, call_data, lang):
        self[0].tag_name = self.get_field_value('inner_tag')
        if self.get_field_value('style'):
            self[0].update_attribs({"style":self.get_field_value('style')})
        self[0][0] = self.get_field_value("set_html")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div>  <!-- with div tag, or any other specified and the set style -->
    <!-- set with unescaped html -->
  </div>
</div>"""


class DivHTML(Widget):

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'set_html':FieldArg("text", "", jsonset=True),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'drag':FieldArg("text", '', valdt=True, jsonset=True),
                        'drop':FieldArg("text", '', valdt=True, jsonset=True),
                        'dropident':FieldArg("url", "")}

    def __init__(self, name=None, brief='', **field_args):
        """A div, containing a string, which will be set as html, without escaping
           drag - if given this is text data sent with a JSON call when this item is dropped
                  if nothing set here, this item will not be set as draggable.
           drop - if given this is text data sent with a JSON call when another item is dropped here
                  if nothing set here, this item will not be set as droppable.
           dropident - ident or label of target which returns a JSON page, called when a drop occurs here"""

        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = ""  # where the html string is to be set
        self.htmlescaped = False
        self.linebreaks=False
        self._dropurl = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self[0] = self.get_field_value("set_html")
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        # dropurl
        self._dropurl = skiboot.get_url(self.get_field_value("dropident"), proj_ident=page.proj_ident)
        # drag
        drag = self.get_field_value("drag")
        if drag:
            self.update_attribs(
{"draggable":"true",
"ondragstart":"SKIPOLE.widgets['{ident}'].dragstartfunc(event, '{data}')".format(ident = self.get_id(),
                                                                                 data = drag)})
        # drop
        drop = self.get_field_value("drop")
        if drop:
            self.update_attribs(
{"ondrop":"SKIPOLE.widgets['{ident}'].dropfunc(event, '{data}')".format(ident = self.get_id(), data = drop),
"ondragover":"SKIPOLE.widgets['{ident}'].allowdropfunc(event)".format(ident = self.get_id())})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets drop and drag"""
        if self._dropurl:
            return self._make_fieldvalues(dropurl=self._dropurl)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
       <!-- attribute style=display:none if hide is True, with drag and drop events if enabled -->
    <!-- set with the set_html string -->
</div>"""


class PreText(Widget):
    """A pre tag, containing text, if an error is raised, then the text is changed
       for the error_message"""

    error_location = 0

    arg_descriptions = {'pre_text':FieldArg("text", "", jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        """
        pre_text: The text appearing in the pre tag
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="pre", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self[0] = self.get_field_value("pre_text")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<pre>  <!-- with widget id and class widget_class -->
    <!-- set with text, replaced by error message on error -->
</pre>"""


class SpanText(Widget):
    """A span tag, containing text"""

     # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'span_text':FieldArg("text", "", jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        """
        span_text: The text appearing in the span tag
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self[0] = self.get_field_value("span_text")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
    <!-- set with text -->
</span>"""


class TagText(Widget):
    """A tag, which can be specified, containing text"""

     # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'tag':FieldArg("text", 'div'),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'tag_text':FieldArg("text", "", jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        """
        tag_text: The text appearing in the tag
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self.tag_name = self.get_field_value('tag')
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        self[0] = self.get_field_value("tag_text")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class and with div tag, or any other specified -->
               <!-- and attribute style=display:none if hide is True -->
    <!-- set with text -->
</div>"""


class ParaText(Widget):
    """A p tag, containing text"""

     # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'para_text':FieldArg("text", "", jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        """
        para_text: The text appearing in the p tag
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="p", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        self[0] = self.get_field_value("para_text")

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<p>  <!-- with widget id and class widget_class -->
    <!-- set with text -->
</p>"""


class DivPara(Widget):
    """A div containing a paragraph, containing text, if an error is raised, then the text is changed
       for the error_message and the error_class given to the paragraph"""

    error_location = (0,0)

    arg_descriptions = {'para_text':FieldArg("text", "", jsonset=True),
                        'para_class':FieldArg("cssclass", ""),
                        'error_class':FieldArg("cssclass", ""),
                        'pre_line':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        para_text: The text appearing in the paragraph
        para_class: The class of the paragraph
        error_class: The class of the error text - which provides the appearance via CSS
                     replaces para_class on error.
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='p')
        # do not insert <br /> - leave that to pre_line
        self[0].linebreaks=False
        self[0][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("pre_line"):
            self[0].attribs={"style":"white-space: pre-line;"}
        if self.get_field_value('para_class'):
            self[0].update_attribs({"class":self.get_field_value('para_class')})
        # self[0][0] could be set by an error message
        if not self.error_status:
            self[0][0] = self.get_field_value("para_text")
        if self.error_status and self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})

            
    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets fieldvalues"""
        return self._make_fieldvalues('para_class', 'error_class')


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
    <p style = "white-space: pre-line;">
    <!-- with class para_class, and style set if pre_line is True -->
    <!-- and class changed to error_class on error -->
        <!-- set with para_text, replaced by error message on error -->
    </p>
</div>"""


class JSONTextLink(Widget):
    """A div, containing an anchor link (used as a button) followed by a paragraph containing text
       If the button is pressed, and the text is visible - the text is hidden, and button text set to 'button_show_text'.
       If the button is pressed, and the text is hidden - a call to a json file is made which should return the
       {(widgname,'para_text'):'new text to place in paragraph', (widgname,'hide'):False} to display the text and
       to set the button to 'button_hide_text'."""

    error_location = (1,0) # set error in para_text

    arg_descriptions = {'para_text':FieldArg("text", "",jsonset=True),
                        'para_class':FieldArg("cssclass", ""),
                        'button_show_text':FieldArg("text", "Show"),
                        'button_hide_text':FieldArg("text", "Hide"),
                        'button_class':FieldArg("cssclass", ""),
                        'json_ident':FieldArg("url", ''),
                        'link_ident':FieldArg("url", 'no_javascript'),
                        'get_field':FieldArg("text", "", valdt=True),
                        'hide':FieldArg("boolean", True, jsonset=True),
                        'pre_line':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        para_text: The text appearing in the paragraph
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='a', attribs={'role':'button'})
        self[0][0] = ""  # where button text is to be set
        self[1] = tag.Part(tag_name="p")
        # do not insert <br /> - leave that to pre_line
        self[1].linebreaks=False
        self[1][0] =  ""  # where para_text is to be set
        self._jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value('button_class'):
             self[0].update_attribs({"class":self.get_field_value('button_class')})
        self[0].update_attribs({"href":"#"})
        if not self.get_field_value("link_ident"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No link ident"
            return
        url = skiboot.get_url(self.get_field_value("link_ident"), proj_ident=page.proj_ident)
        if url:
            get_fields = {self.get_formname("get_field"):self.get_field_value("get_field")}
            url = self.make_get_url(page, url, get_fields, force_ident=True)
            self[0].update_attribs({"href": url})
        else:
            self._error = "Warning: Invalid link ident"
            return

        para_style = {}

        if self.get_field_value("para_class"):
            para_style["class"] = self.get_field_value("para_class")

        if self.get_field_value("hide") and (not self.error_status):
            self[0][0] = self.get_field_value("button_show_text")
            if self.get_field_value("pre_line"):
                para_style["style"] = "display:none;white-space: pre-line;"
            else:
                para_style["style"] = "display:none;"
        else:
            self[0][0] = self.get_field_value("button_hide_text")
            if self.get_field_value("pre_line"):
                para_style["style"] = "white-space: pre-line;"

        if para_style:
            self[1].attribs=para_style

        if not self.error_status:
            # only set para_text if an error is not already there
            self[1][0] = self.get_field_value("para_text")
        if not self.get_field_value("json_ident"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No link ident to JSON"
            return
        self._jsonurl = skiboot.get_url(self.get_field_value("json_ident"),  proj_ident=page.proj_ident)
        if not self._jsonurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler"""
        if not self._jsonurl:
            return
        jscript = """  $("#{ident} a").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        return jscript + self._make_fieldvalues( 'button_show_text', 'button_hide_text', url=self._jsonurl)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <a href="#" role="button">  <!-- with class button_class -->
    <!-- set with button_show_text or button_hide_text -->
  </a>
  <p style = "white-space: pre-line;">  <!-- Either hidden or not, depending on hide -->
    <!-- with class para_class, and style set if pre_line is True -->
    <!-- para_text or error message -->
  </p>
</div>"""



class JSONDivLink(Widget):
    """A div, containing an anchor link (used as a button) followed by a div containing text or html content
       If the button is pressed, and the content is visible - the content is hidden, and button text set to 'button_show_text'.
       If the button is pressed, and the content is hidden - a call to a json file is made which should return the
       {(widgname,'div_content'):'new content to place in div', (widgname,'hide'):False} to display the content and
       to set the button to 'button_hide_text'."""

    error_location = (1,0) # set error in div_content

    arg_descriptions = {'div_content':FieldArg("text", "",jsonset=True),
                        'div_class':FieldArg("cssclass", ""),
                        'button_show_text':FieldArg("text", "Show"),
                        'button_hide_text':FieldArg("text", "Hide"),
                        'button_class':FieldArg("cssclass", ""),
                        'json_ident':FieldArg("url", ''),
                        'link_ident':FieldArg("url", 'no_javascript'),
                        'get_field':FieldArg("text", "", valdt=True),
                        'hide':FieldArg("boolean", True, jsonset=True),
                        'htmlescaped':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        div_content: The content appearing in the div
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='a', attribs={'role':'button'})
        self[0][0] = ""  # where button text is to be set
        self[1] = tag.Part(tag_name='div')
        self[1][0] =  ""  # where div_content is to be set
        self._jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value('button_class'):
             self[0].update_attribs({"class":self.get_field_value('button_class')})
        self[0].update_attribs({"href":"#"})
        if self.get_field_value("link_ident"):
            url = skiboot.get_url(self.get_field_value("link_ident"), proj_ident=page.proj_ident)
            if url:
                get_fields = {self.get_formname("get_field"):self.get_field_value("get_field")}
                url = self.make_get_url(page, url, get_fields, force_ident=True)
                self[0].update_attribs({"href": url})

        if self.get_field_value("hide") and (not self.error_status):
            self[1].attribs={"style":"display:none;"}
            self[0][0] = self.get_field_value("button_show_text")
        else:
            self[1].attribs={}
            self[0][0] = self.get_field_value("button_hide_text")
        if self.get_field_value('div_class'):
            self[1].update_attribs({"class":self.get_field_value('div_class')})
        if not self.error_status:
            # only set div_content if an error is not already there
            self[1][0] = self.get_field_value("div_content")
        if not self.get_field_value("htmlescaped"):
            # set the div to not escape its contents
            self[1].htmlescaped = False
        # set an id in the button and div
        self[0].insert_id()
        self[1].insert_id()
        if not self.get_field_value("json_ident"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No link ident to JSON"
            return
        self._jsonurl = skiboot.get_url(self.get_field_value("json_ident"),  proj_ident=page.proj_ident)
        if not self._jsonurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler"""
        jscript = """  $("#{buttonident}").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(
        ident = self.get_id(),
        buttonident = self[0].get_id(), # ident of the button
        )
        return jscript + self._make_fieldvalues("button_hide_text", "button_show_text", "get_field",
                                                htmlescaped = "text" if self.get_field_value("htmlescaped") else "html",
                                                divident = self[1].get_id(),
                                                buttonident = self[0].get_id(),
                                                url = self._jsonurl
                                                )

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <a href="#" role="button">  <!-- with class button_class -->
    <!-- set with button_show_text or button_hide_text -->
  </a>
  <div>  <!-- with class div_class, and either hidden or not, depending on hide -->
    <!-- div_content shown as either text or html -->
  </div>
</div>"""


class TextBlockPara(Widget):
    """A paragraph, containing a TextBlock, if a error is raised, then the text is changed
       for the error_message and the error_class given to the paragraph"""

    error_location = 0

    arg_descriptions = {'textblock_ref':FieldArg("textblock_ref", ""),
                        'textblock_project':FieldArg("text", ""),
                        'text_refnotfound':FieldArg("text", ""),
                        'text_replaceblock':FieldArg("text", "" ,jsonset=True),
                        'replace_strings':FieldArgList("text"),
                        'linebreaks':FieldArg("boolean", True),
                        'error_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        textblock_ref: The reference of the TextBlock appearing in the paragraph
        textblock_project: Set with a project name if the TextBlock is defined in a sub project
        text_refnotfound: text to appear if the textblock is not found
        text_replaceblock: text set here will replace the textblock
        replace_strings: A list of strings, if given, will be used with python % operator on the textblock text (not on text_replaceblock)
        linebreaks: Set True if linebreaks in the text are to be shown as html breaks
        error_class: The class of the error text - which provides the appearance via CSS
                     replaces widget_class on error.
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="p", brief=brief, **field_args)
        self[0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        if not self.error_status:
            linebreaks = bool(self.get_field_value('linebreaks'))
            if self.get_field_value('text_replaceblock'):
                # no textblock, just the replacement text
                self[0] = self.get_field_value('text_replaceblock')
                if not linebreaks:
                    self.linebreaks = False             
            else:
                # define the textblock
                tblock = self.get_field_value("textblock_ref")
                tblock.project = self.get_field_value('textblock_project')
                tblock.failmessage = self.get_field_value('text_refnotfound')
                tblock.linebreaks = linebreaks
                if self.get_field_value('replace_strings'):
                    tblock.replace_strings = self.get_field_value('replace_strings')
                # place it at location 0
                self[0] = tblock
        if self.error_status and self.get_field_value('error_class'):
            self.update_attribs({"class":self.get_field_value('error_class')})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets fieldvalues"""
        return self._make_fieldvalues('widget_class','error_class','linebreaks')

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<p>  <!-- with widget id, and class widget_class replaced by error_class on failure -->
   <!-- set with either text_replaceblock or textblock content, replaced by error message on error -->
</p>"""



class TextBlockDiv(Widget):
    """A div, containing a TextBlock. The TextBlock is not escaped, so may contain
       html which is set directly into the div"""

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {'textblock_ref':FieldArg("textblock_ref", ""),
                        'textblock_project':FieldArg("text", ""),
                        'content_refnotfound':FieldArg("text", ""),
                        'content_replaceblock':FieldArg("text", "" ,jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        textblock_ref: The reference of the TextBlock appearing in the div
        textblock_project: Set with a project name if the TextBlock is defined in a sub project
        content_refnotfound: content to appear if the textblock is not found
        content_replaceblock: content set here will replace the textblock
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = ''
        self.htmlescaped = False
        self.linebreaks = False

    def _build(self, page, ident_list, environ, call_data, lang):
        # define the textblock
        tblock = self.get_field_value("textblock_ref")
        tblock.project = self.get_field_value('textblock_project')
        tblock.failmessage = self.get_field_value('content_refnotfound')
        tblock.escape = False
        tblock.linebreaks = False
        # place it at location 0
        if self.get_field_value("content_replaceblock"):
            self[0] = self.get_field_value("content_replaceblock")
        else:
            self[0] = tblock

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id, and class widget_class -->
   <!-- set with either content_replaceblock or textblock content as unescaped html -->
</div>"""


class ShowPara1(Widget):
    """A div, containing a paragraph of text followed by an error div and paragraph.
    The text paragraph is normally visible (depending on the show_para field)
    The error div is invisible, and becomes visible on error."""

    error_location = (1,0,0)

    arg_descriptions = {'para_text':FieldArg("text", default="", valdt=False, jsonset=True),
                        'show_para':FieldArg("boolean", default=True, valdt=False, jsonset=True),
                        'para_class':FieldArg("cssclass", ""),
                        'error_class':FieldArg("cssclass", ""),
                        'pre_line':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        para_text: The text appearing in the paragraph
        show_para: True if the paragraph is to be visible, False if not
        para_class: The class of the paragraph - which provides the appearance via CSS
        error_class: The class of the error text - which provides the appearance via CSS
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # The para_text
        self[0] = tag.Part(tag_name="p")
        # do not insert <br /> - leave that to pre_line
        self[0].linebreaks=False
        self[1] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[1][0] = tag.Part(tag_name="p")
        self[1][0][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        para_style = {}

        if self.get_field_value("para_class"):
            para_style["class"] = self.get_field_value("para_class")

        if self.get_field_value("show_para"):
            if self.get_field_value("pre_line"):
                para_style["style"] = "white-space: pre-line;"
        else:
            if self.get_field_value("pre_line"):
                para_style["style"] = "display:none;white-space: pre-line;"
            else:
                para_style["style"] = "display:none;"

        if para_style:
            self[0].attribs=para_style

        self[0][0] = self.get_field_value("para_text")
        if self.error_status:
            self[1].del_one_attrib("style")
        if self.get_field_value('error_class'):
            self[1].update_attribs({"class":self.get_field_value('error_class')})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <p style = "white-space: pre-line;">
    <!-- with class para_class, and style set if pre_line is True -->
    <!-- hidden if show_para is False -->
    <!-- set with para_text -->
  </p>
  <div>  <!-- with class error_class, normally hidden, shown on error -->
    <p> <!-- set with error message --> </p>
  </div>
</div>"""


class ShowPara2(Widget):
    """A div, containing an x button which on pressed will hide it, and after the button, another div holding text."""

     # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'para_text':FieldArg("text", default="", valdt=False, jsonset=True),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'inner_class':FieldArg("cssclass", ""),
                        'close_class':FieldArg("cssclass", ""),
                        'close_style':FieldArg("cssstyle", ""),
                        'pre_line':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='div', attribs={"onclick":"this.parentElement.style.display='none'"})
        self[0][0] = tag.HTMLSymbol("&times;")
        self[1] = tag.Part(tag_name='div')
        self[1][0] = tag.Part(tag_name='p')
        # do not insert <br /> - leave that to pre_line
        self[1][0].linebreaks=False
        self[1][0][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        # define the text paragraph
        self[1][0][0] = self.get_field_value("para_text")
        if self.get_field_value("inner_class"):
            self[1].update_attribs({"class":self.get_field_value("inner_class")})
        if self.get_field_value("close_class"):
            self[0].update_attribs({"class":self.get_field_value("close_class")})
        if self.get_field_value("close_style"):
            self[0].update_attribs({"style":self.get_field_value("close_style")})
        if self.get_field_value("pre_line"):
            self[1][0].attribs = {"style":"white-space: pre-line;"}

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div onclick="this.parentElement.style.display='none'"> <!-- with class close_class and style close_style -->
    &times;
  </div>
  <div> <!-- with class inner_class -->
    <p style = "white-space: pre-line;"> <!-- style set if pre_line is True -->
      <!-- para_text -->
    </p>
  </div>
</div>"""



